"""
A scraper to collect image plane movies, stills, and sequences.
"""

import os
import re
import pymel.core as pm
 
IMAGE_PLANE_FILENAME_REGEX = re.compile(r"^(.*)\.(\d+)\.([a-z0-9]+$)")
 

def run(_):
    """
    Find image_plane sequences.
    """
    result = []

    for ip in pm.ls(type="imagePlane"):
        iptype = ip.attr("type").get()
        if iptype == 1: #texture
            continue
        plug = ip.attr("imageName")
        path = plug.get().strip()
        if path:
            # movie or still or sequence
            if iptype == 0 and  ip.attr("useFrameExtension").get():
                #sequence - replace frameExt with *
                match = IMAGE_PLANE_FILENAME_REGEX.match(path)
                if match:
                    root, _, ext = match.groups()
                    path = "{}.*.{}".format(root, ext)

            result.append({
                "plug":plug.name(),
                "path":path
            })

    return result



"""
A scraper to collect image plane movies, stills, and sequences.

It demonstrates how to use attr.get(time=t) to evaluate a sequence for a frame
range without evaluating the whole DG. It means that if an image plane uses 10
files from a folder of 5000, we only upload the 10 files.
"""

import re
import glob
import pymel.core as pm
from ciocore.sequence import Sequence
from ciomaya.lib import scraper_utils

IMAGE_PLANE_FILENAME_REGEX = re.compile(r"^(.*)\.(\d+)\.([a-z0-9]+$)")

def run(node):
    """
    Find image_plane sequences.
    """

    paths = []
    sequence = Sequence.create(pm.PyNode(node).attr("frameSpec").get())

    for image_plane in pm.ls(type="imagePlane"):
        paths.extend(paths_for_image_plane(image_plane, sequence))
    
    paths = scraper_utils.expand_workspace(paths)

    return paths


def paths_for_image_plane(image_plane, sequence):
    """
    image_plane type may be image (possibly sequence), procedural textur, or movie

    If it's a movie, or sequence is off, return the path.
    For textures, there's no file so return empty list.

    For sequences,
    1. evaluate the imagePlane.frameExtension for the conductorRender node
       range.
    2. Determine padding from the sequence on disk.
    3. Compute the filenames we need for the rendered range.
    """

    ws = pm.Workspace()
    iptype = image_plane.attr("type").get()
    if iptype == 1:  # texture
        return []

    plug = image_plane.attr("imageName")
    plug_name = plug.name()
    path = ws.expandName(plug.get().strip())

    match = IMAGE_PLANE_FILENAME_REGEX.match(path)

    if iptype == 2 or (not image_plane.attr("useFrameExtension").get()) or (not match):
        # not sequence
        return [{
            "plug": plug_name,
            "path": path
        }]

    root, _, ext = match.groups()
    existing = glob.glob("{}.*.{}".format(root, ext))
    if not existing:
        return []
    padding = min([len(fn.split(".")[-2]) for fn in existing])

    # By evaluating, we account for expressions. offsets, driven keys, timewarps etc.
    frame_seq =  Sequence.create([image_plane.attr(
        "frameExtension").get(time=f) for f in sequence])
    path = "{}.<f{}>.{}".format(root, padding, ext)
    pathnames = scraper_utils.resolve_to_sequence(path, frame_seq)
    return [{"plug": plug_name, "path": p} for p in pathnames ]




"""
A basic scraper to collect paths from Bifrost nodes.
"""

import os
import pymel.core as pm
from ciomaya.lib import scraper_utils
from ciocore.sequence import Sequence

################## LEAVE HERE ###################
# Investigate how these attributes work before implementing them.
# ATTRS = {
    # "bifrostGraph": {
    #     "bifShape": ["aiFilename"]
    # },
    # "bifmeshio": { "BifMeshImportNode": ["bifMeshDirectory"] }
    # "Boss": {
    #     "BossEXRInfluence": ["exrFilename"],
    #     "BossGeoProperties": ["cacheName", "cacheFolder"],
    #     "BossSpectralWave": ["velocityCacheName", "cacheName", "cacheFolder", "foamCacheName"],
    #     "BossWaveSolver": ["velocityCacheName", "cacheName", "foamCacheName", "cacheFolder", "remappedInputCacheName"]
    # }
# }

def run(node):
    """
    Find paths in Bifrost nodes.
    """

    paths = scraper_utils.get_paths({"bifmeshio": { "BifMeshImportNode": ["bifMeshDirectory"] }})

    # we want to add all files in the directory for an explicit list of frames,
    # which we evaluate from the time attribute in the  BifMeshImportNode node.
    result = []
    for path in paths:
        cache_node =  pm.Attribute(path["plug"]).node()
        time_plug = cache_node.attr("time")
        start_time_plug = cache_node.attr("startFrame")
        end_time_plug = cache_node.attr("endFrame")
        frames = [int(round(time_plug.get(time=x))) for x in Sequence.create(pm.PyNode(node).attr("frameSpec").get())]
        seq = Sequence.create(frames).intersection(Sequence.create(start_time_plug.get(),end_time_plug.get()))
        template = os.path.join(path["path"], "*.{frame:04d}.bif")
        result.extend([{"path": p, "plug": path["plug"]} for p in seq.expand_format(template)])

    result = scraper_utils.expand_workspace(result)

    return result



"""
A scraper to return paths used by XGen.

This scraper searches description files contained in xgmPalette files.
"""
import maya.api.OpenMaya as om
import pymel.core as pm
import os


def run(_):
    """
    Find paths related to xgen.

    This is currently a stub.
    """
    result = []
    scene_dir = os.path.dirname(pm.sceneName())
    for node in pm.ls(type="xgmPalette"):
        for att in ["xgFileName", "xgBaseFile", "xgDeltaFiles"]:
            attr = node.attr(att)
            path = attr.get()
            if path:
                path = path if os.path.isabs(path) else os.path.join(scene_dir, path)
                result.append({"path": path, "plug":attr.name()})
    return result
 


import abc
import importlib
import inspect
import os
import re
import pymel.core as pm
from ciocore.gpath import Path
from ciocore.gpath_list import GLOBBABLE_REGEX, PathList
from ciocore.validator import Validator, ValidationError


from ciomaya.lib import const as k
from ciomaya.lib import layer_utils
from ciomaya.lib import asset_cache
from ciomaya.lib import software

RENDERER_PACKAGE_REQUIREMENTS = {
    "arnold": "arnold-maya",
    "vray": "v-ray-maya",
    "renderman": "renderman-maya",
    "redshift": "redshift-maya",
}

WARNING_ICON = "ConductorWarning_18x18.png"
INFO_ICON = "ConductorInfo_18x18.png"


class ValidateCamera(Validator):
    def run(self, _):
        if not any(cam.attr("renderable").get() for cam in pm.ls(type="camera")):
            self.add_warning(
                "No renderable cameras. You may want to make at least one camera renderable in Render Settings.")


class ValidateUploadDaemon(Validator):
    def run(self, _):
        use_daemon = self._submitter.attr("useUploadDaemon").get()
        if not use_daemon:
            return
        module_path = pm.moduleInfo(path=True, moduleName="conductor")
        conductor_executable = os.path.join(
            os.path.dirname(module_path), "bin", "conductor")
        msg = 'This submission expects an uploader daemon to be running.\n After you press submit you can open a shell and enter: "{}" uploader'.format(
            conductor_executable)

        location = (self._submitter.attr("locationTag").get() or "").strip()
        if location:
            msg = 'This submission expects an uploader daemon to be running and set to a specific location tag.\nAfter you press submit you can open a shell and type: "{}" uploader --location {}'.format(
                conductor_executable,
                location)
        msg += "\nCheck the script editor for details"
        self.add_notice(msg)
        # By also printing the message, the user can copy paste
        # `/path/to/conductor uploader --location blah` from the console.
        print msg


class ValidateTaskCount(Validator):
    def run(self, _):
        count = self._submitter.attr("taskCount").get()
        if count > 1000:
            self.add_notice(
                "This submission contains over 1000 tasks ({}). Are you sure this is correct?".format(count))


class ValidateSelectedRenderer(Validator):
    def run(self, layername):

        current_renderer = pm.PyNode(
            "defaultRenderGlobals").attr("currentRenderer").get()
        try:
            required_package = RENDERER_PACKAGE_REQUIREMENTS[current_renderer]
        except KeyError:
            return

        self._submitter.attr("pluginSoftware").get()
        if required_package not in [p.split(" ")[0] for p in self._submitter.attr("pluginSoftware").get() if p]:
            self.add_warning(
                "The renderer for layer '{}' is set to '{}' but no versions of the plugin software '{}' are selected in the submitter. Are you sure this is correct?".format(layername, current_renderer, required_package))


class ValidateArnoldTiledTextures(Validator):
    def run(self, _):

        if not pm.PyNode("defaultRenderGlobals").attr("currentRenderer").get() == "arnold":
            return
        try:
            render_options = pm.PyNode("defaultArnoldRenderOptions")
        except pm.MayaNodeError:
            self.add_warning(
                "Current renderer is set to Arnold, but there's no defaultArnoldRenderOptions node")
            return

        msg = "It is recommended to generate Arnold Tiled Textures (TX files) locally as they can't be generated efficiently on the render nodes.\n"
        msg += "Use the Arnold Texture Manager to generate tx files, then set the following attributes in the Arnold tab of Render Settings.\n"
        msg += "Switch Auto Convert Textures to Off\n"
        msg += "Switch Use Existing TX Textures to On\n"

        auto_tx = render_options.attr("autotx").get()
        use_existing = render_options.attr("use_existing_tiled_textures").get()

        if auto_tx or not use_existing:
            self.add_warning(msg)


class ValidateArnoldRenderOptions(Validator):
    def run(self, _):
        if not pm.PyNode("defaultRenderGlobals").attr("currentRenderer").get() == "arnold":
            return

        try:
            render_options = pm.PyNode("defaultArnoldRenderOptions")
        except pm.MayaNodeError:
            self.add_warning(
                "Current renderer is set to Arnold, but there's no defaultArnoldRenderOptions node")
            return

        if not render_options.attr("threads_autodetect").get():
            self.add_warning(
                "Autodetect-Threads is turned off which could cause suboptimal machine usage and incur unnecessary costs. You may want to switch it back on in the Render Settings window, System tab.")

        if render_options.attr("denoiseBeauty").get():
            self.add_warning(
                "You have the Optix denoiser turned on. This will not work unless you also select a machine with a capable GPU. Also note, the Optix denoiser is not temporally coherent. To turn off Optix denoiser, enter the following in the MEL script editor: setAttr \"defaultArnoldRenderOptions.denoiseBeauty\" 0;")


class ValidateMissingAssets(Validator):

    def run(self, _):

        path_list = PathList()

        for gpath in asset_cache.data(self._submitter):

            # we can assunme that all globbable paths, (those containing glob
            # characters) cannot possibly represent missing files because they are
            # literally resolved  by checking what matches on disk. Ignore them.
            if not GLOBBABLE_REGEX.search(gpath.posix_path()):
                path_list.add(gpath)

        # path_list has taken care of deduplication
        missing = []
        for gpath in path_list:
            pp = gpath.posix_path()
            if not os.path.exists(pp):
                missing.append(pp)

        if missing:
            self.add_warning(
                "Some assets do not exist on disk. See the script editor for details. You can continue if you don't need them.")

            print "----- Conductor Asset Validation -------"
            for asset in missing:
                print "Missing: {}".format(asset)


class ValidateDestinationDirectory(Validator):
    def run(self, _):
        dest_directory = Path(self._submitter.attr(
            "destinationDirectory").get()).posix_path(with_drive=False)
        for gpath in asset_cache.data(self._submitter):
            asset_path = gpath.posix_path(with_drive=False)
            if asset_path.startswith(dest_directory):
                print "Some of your upload assets exist in the specified output destination directory\n. {} contains {}".format(
                    dest_directory, asset_path)
                self.add_error(
                    "The destination directory for output files contains assets that are in the upload list. This will cause your render to fail. See the script editor for details.")
                break
            if dest_directory.startswith(asset_path):
                print "You are trying to upload a directory that contains your destination directory.\n. {} contains {}".format(
                    asset_path, dest_directory)
                self.add_error(
                    "One of your assets is a directory that contains the specified output destination directory. This will cause your render to fail. See the script editor for details.")
                break


class ValidateGeneralRenderSettings(Validator):
    def run(self, _):
        if pm.PyNode("defaultRenderGlobals").attr("modifyExtension").get():
            self.add_warning(
                "You have 'renumber frames' turned on in render settings. All your files will be renamed to frame 1. We strongly recommend that you turn this setting OFF.")


class ValidateYeti(Validator):
    def run(self, _):
        # 
        if not (software.detect_yeti() and pm.about(ntOS=True)  and pm.ls(type="pgYetiMaya")  ):
            return
 
        msg = """You have the YETI plugin loaded and Yeti nodes in your scene.

Since Conductor render nodes run on Linux, you must ensure that every Yeti asset
can be found on the Linux filesystem. To do this, you should define the
PG_IMAGE_PATH variable and make sure it contains an entry for every directory
where your assets are, but with the drive letter removed and backslashes
replaced with forward slashes. Specifically, in the Submitter UI, open the Extra
Environment section and add an entry for each directory like the following
example:

Suppose you have these assets 
C:\\Users\\roberto\\textures\\texture.1.tif
C:\\Users\\roberto\\textures\\texture.2.tif
W:\\Production\\assets\\textures\\texture.3.tif

Then add these extra environment entries.

PG_IMAGE_PATH /Users/roberto/textures
PG_IMAGE_PATH /Production/assets/textures
"""
        self.add_notice("{}\nThis information has been printed to the script editor.".format(msg))

        print(msg)


# Implement more validators here
####################################
####################################


def run(node, dry_run=False):

    errors, warnings, notices = _run_validators(node)

    if errors:
        for error in errors:
            pm.displayError(error)
        msg = "There are some critical issues. See the script editor for details."
        raise ValidationError(msg)
    if notices or warnings:
        dialog_result = pm.layoutDialog(
            ui=pm.Callback(result_window, warnings, notices, dry_run), title="Validation")
        if not dialog_result == "okay":
            msg = "Submission cancelled by user."
            raise ValidationError(msg)


def _run_validators(node):

    layer_policy = node.attr("renderLayers").get()
    validators = [plugin(node) for plugin in Validator.plugins()]
    if layer_policy == k.CURRENT_LAYER:
        layers = [pm.editRenderLayerGlobals(q=True, currentRenderLayer=True)]
    else:
        layers = layer_utils.get_renderable_legacy_layers()
    for layer in layers:
        layername = layer_utils.get_layer_name(layer)
        with layer_utils.layer_context(layer):
            for validator in validators:
                validator.run(layername)

    errors = list(set.union(*[validator.errors for validator in validators]))
    warnings = list(
        set.union(*[validator.warnings for validator in validators]))
    notices = list(set.union(*[validator.notices for validator in validators]))
    return errors, warnings, notices


def result_window(warnings, notices, dry_run):
    form = pm.setParent(q=True)
    text = pm.text(
        label="Please read the notices before you continue!")

    if not dry_run:
        cancel_button = pm.button(label="Cancel", command=pm.Callback(
            pm.layoutDialog, dismiss="abort"))

    okay_label = "Close" if dry_run else "Submit"

    okay_button = pm.button(label=okay_label, command=pm.Callback(
        pm.layoutDialog, dismiss="okay"))

    scroll = pm.scrollLayout(bv=True)
    pm.setParent("..")

    pm.formLayout(form, edit=True, width=600)
    form.attachForm(text, "left", 2)
    form.attachForm(text, "right", 2)
    form.attachForm(text, "top", 2)
    form.attachNone(text, "bottom")

    form.attachForm(scroll, "left", 2)
    form.attachForm(scroll, "right", 2)
    form.attachControl(scroll, "top", 2, text)
    form.attachControl(scroll, "bottom", 2, okay_button)

    if not dry_run:
        form.attachForm(cancel_button, "left", 2)
        form.attachNone(cancel_button, "top")
        form.attachForm(cancel_button, "bottom", 2)
        form.attachPosition(cancel_button, "right", 2, 50)

    if not dry_run:
        form.attachPosition(okay_button, "left", 2, 50)
    else:
        form.attachForm(okay_button, "left", 2)

    form.attachForm(okay_button, "right", 2)
    form.attachNone(okay_button, "top")
    form.attachForm(okay_button, "bottom", 2)

    pm.setParent(scroll)
    col = pm.columnLayout(adj=True)

    for warning in warnings:
        _create_notice_widget(col, warning, WARNING_ICON)
        pm.setParent(col)
        pm.separator(height=8, style="in")
    for notice in notices:
        _create_notice_widget(col, notice, INFO_ICON)
        pm.setParent(col)
        pm.separator(height=8, style="in")
    pm.setParent(form)


def _create_notice_widget(column, notice, image):

    pm.setParent(column)

    form = pm.formLayout(nd=100, width=600)
    icon = pm.iconTextStaticLabel(style="iconOnly", image1=image)
    text = pm.text(label=notice.strip(), ww=True, align="left")

    form.attachForm(icon, "left", 2)
    form.attachNone(icon, "right")
    form.attachForm(icon, "top", 4)
    form.attachForm(icon, "bottom", 4)

    form.attachControl(text, "left", 10, icon)
    form.attachForm(text, "right", 2)
    form.attachForm(text, "top", 4)
    form.attachForm(text, "bottom", 4)



"""
Handle the UI for extra assets:
"""

import pymel.core as pm
from ciomaya.lib import const as k
from ciomaya.lib.ae import AEcommon


def create_ui(node_attr):
    """Build static UI"""
 
    with AEcommon.ae_template():
        grp = pm.textFieldGrp("taskTemplateGrp", label="Task Template")
        label = pm.layout(grp, q=True, childArray=True)[0]
        pm.popupMenu(parent=label)
        pm.menuItem(label="Reset")
        populate_ui(node_attr)

def populate_ui(node_attr):
    """Populate / reconfigure UI for the current node"""
    attr = pm.Attribute(node_attr)
    widgets = _get_widgets()
    # index 2 is the field. (1 is the label) 
    pm.connectControl(widgets["field"], attr, index=2)
    # reconfigure the popup for this node.
    item = pm.popupMenu(widgets["popup_menu"], q=True, itemArray=True)[0]
    pm.menuItem(item, edit=True, command=pm.Callback(_on_reset, attr))

def _get_widgets(parent=None):
    if not parent:
        parent = pm.setParent(q=True)
    field_grp = AEcommon.find_ui("taskTemplateGrp", parent)
    label = pm.layout(field_grp, q=True, childArray=True)[0]

    return {
        "field": field_grp,
        "popup_menu": pm.control( label, q=True, popupMenuArray=True)[0]
    }

def _on_reset(attribute):
    attribute.set(k.DEFAULT_TEMPLATE)
 


"""
Handle the UI for custom frames and use scout frames:

In both cases, when the use<Attrname> checkbox is unchecked,
the field is hidden.

"""

import pymel.core as pm
from ciomaya.lib import const as k
from ciomaya.lib.ae import AEcommon
import random

constants = {
    "customRange": {
        "text_label": "Custom Range",
        "annotation": "Specifies a set of frames to render. Provide a comma-separated list of progressions. Example 1,7,10-20,30-60x3,1001.",
        "bool_attr": "useCustomRange"
    },
    "scoutFrames": {
        "text_label": "Scout Frames",
        "annotation": "Specifies a set of frames to test render on Conductor before other frames. Provide a comma-separated list of progressions. Example 1,7,10-20,30-60x3,1001.",
        "bool_attr": "useScoutFrames"
    },
}


def create_ui(node_attr):
    """Build static UI.

    """
    attr = pm.Attribute(node_attr)
    key = attr.attrName(True)

    with AEcommon.ae_template():
        tflabel = constants[key]["text_label"]
        cblabel = "Use {}".format(tflabel)
        pm.checkBoxGrp("frameSpecCheckbox", label=cblabel)

        pm.frameLayout("frameSpecFrame",
                       visible=False, lv=False, cl=False, cll=False
                       )

        pm.textFieldGrp("frameSpecField", label=tflabel)

        pm.setParent("..")  # out of frameLayout
        populate_ui(node_attr)


def populate_ui(node_attr):
    """Reconfigure action buttons when node changes."""

    attr = pm.Attribute(node_attr)
    key = attr.attrName(True)
    widgets = _get_widgets()
    bool_attr = attr.node().attr(constants[key]["bool_attr"])
    bool_val = bool_attr.get()
    val = attr.get()

    pm.checkBoxGrp(
        widgets["checkbox"],
        edit=True,
        value1=bool_val,
        changeCommand=pm.Callback(
            _on_bool_changed, bool_attr, widgets["checkbox"],  widgets["frame"]),
    )

    pm.textFieldGrp(
        widgets["field"],
        edit=True,
        text=val,
        changeCommand=pm.Callback(
            _on_text_changed, attr,  widgets["field"]),
    )

    _on_bool_changed(bool_attr,  widgets["checkbox"], widgets["frame"])


def _get_widgets(parent=None):
    if not parent:
        parent = pm.setParent(q=True)
    return {
        "checkbox": AEcommon.find_ui("frameSpecCheckbox", parent),
        "frame":  AEcommon.find_ui("frameSpecFrame", parent),
        "field":  AEcommon.find_ui("frameSpecField", parent),
    }


def _on_bool_changed(attr, checkbox, frame):
    val = pm.checkBoxGrp(checkbox, q=True, value1=True)
    attr.set(val)
    pm.frameLayout(frame, edit=True, enable=True, visible=val)


def _on_text_changed(attr, widget):
    attr.set(pm.textFieldGrp(widget, q=True, text=True))
    AEcommon.print_setAttr_cmd(attr)



import re

AE_TEXT_WIDTH = 145
AE_SINGLE_WIDTH = 70
AE_TOTAL_WIDTH = AE_TEXT_WIDTH + (AE_SINGLE_WIDTH * 5)
SUPPRESS_EXTRA_ATTS = [
    "useCustomRange",
    "useScoutFrames",
    "doScrape",
    "output",
    "pluginSoftware",
    "instanceTypeName",
    "title",
    "projectName",
    "startFrame",
    "endFrame",
    "byFrame",
    "customRange",
    "scoutFrames",
    "locationTag"
]
USE_FIXTURES = True
FIXTURES_PATH = "/Users/julian/dev/conductor/conductor-maya/src/ciomaya/dev/fixtures"

CURRENT_LAYER = 0
RENDERABLE_LAYERS = 1
CUSTOM = 2

DEFAULT_TEMPLATE = "Render -r <Renderer> -s <start> -e <end> -b <step> -rl \"<RenderLayer>\" -rd \"<OutputPath>\"  -proj \"<WorkspacePath>\" \"<SceneFile>\""

OTHER_TEMPLATES = [
    "Render -r <Renderer>  -ai:lve 3 -s <start> -e <end> -b <step> -rl \"<RenderLayer>\" -rd \"<OutputPath>\"  -proj \"<WorkspacePath>\" \"<SceneFile>\""
]


DEFAULT_TITLE = "Maya:<Renderer> - <Scene> <RenderLayer>"
DEFAULT_AUTOSAVE_TEMPLATE = "cio_<Scene>"

DEFAULT_INSTANCE_TYPE = "n1-standard-4"



"""
A basic scraper to collect paths from Maya filename attributes.

If an attribute is intended to contain a filename, it should have the
usedAsFilename flag set. This scraper returns the value for all such attributes
in the scene, including those contained in 3rd party plugins.
"""

import os
import re

import maya.api.OpenMaya as om
import pymel.core as pm

# nodetypes that can always be safely ignored
NODETYPE_BLACKLIST = [
    "partition",
    "shadingEngine",
    "objectSet",
    "script",
    "objectRenderFilter",
    "objectMultiFilter",
    "objectAttrFilter",
    "objectNameFilter",
    "objectScriptFilter",
    "objectTypeFilter",
    "nodeGraphEditorInfo",
    "renderSettingsCollection",
    "simpleSelector",
    "partition",
    "conductorRender",
    "displayLayer",
    "defaultLightList",
    "renderGlobals",
    "renderLayer",
    "renderQuality",
    "defaultRenderUtilityList",
    "defaultRenderingList",
    "resolution",
    "defaultShaderList",
    "defaultTextureList",
    "dof",
    "dynController",
    "globalCacheControl",
    "hardwareRenderGlobals",
    "hardwareRenderingGlobals",
    "hikSolver",
    "hyperGraphInfo",
    "hyperLayout",
    "ikRPsolver",
    "ikSCsolver",
    "ikSplineSolver",
    "ikSystem",
    "materialInfo",
    "displayLayerManager",
    "lightLinker",
    "lightList",
    "polyMergeEdge",
    "polyMergeFace",
    "poseInterpolatorManager",
    "renderGlobalsList",
    "renderLayerManager",
    "renderSetup",
    "selectionListOperator",
    "sequenceManager",
    "shapeEditorManager",
    "strokeGlobals",
    "multilisterLight",
    "time",
    "aiOptions",
    "hwRenderGlobals",
]


# Don't use fileTextureName
# Use computedFileTextureNamePattern instead
ATTR_BLACKLIST = [
    "file.fileTextureName"
]


def run():
    """
    Find values in attributes that have usedAsFilename flag set.

    Test all nodes in the scene.

    For optimization purposes: As we find nodes with NO usedAsFilename atts, we
    add the node_type_name to a blacklist. Aw we find nodes WITH usedAsFilename atts,
    we add the atts to a whitelist object, keyed on the node_type_name. In this way,
    we can quickly eliminate a node (if it's type is in the blacklist), or know
    which attributes are viable filename candidates.

    """
    blacklist = []
    whitelist = {}
    result = []

    sel = om.MSelectionList()
    # add all nodes in all namespaces
    sel.add("::*")
    num = sel.length()

    for i in xrange(num):
        dn_obj = sel.getDependNode(i)
        fn_node = om.MFnDependencyNode(dn_obj)
        node_type_name = fn_node.typeName
        if node_type_name in blacklist  or node_type_name in NODETYPE_BLACKLIST:
            continue
        if node_type_name not in whitelist:  # yet
            atts = get_potential_filename_atts(fn_node)
            print fn_node.name()
            if atts:
                whitelist[node_type_name] = atts
            else:
                blacklist.append(node_type_name)

        if node_type_name in whitelist:
            attnames = whitelist[node_type_name]

            result += get_filename_paths_for_node(fn_node, attnames)

    for r in result:
        r["path"] = re.sub("<f>", "*", r["path"])

    result = extend_with_tx_paths(result)
    return result


def get_potential_filename_atts(fn_node):
    """
    Get the list of atts for a node that are potentially filenames.

    They either have usedAsFilename set. Or they are string/stringArray and
    their name contains the word file or path.
    """
    result = []
    num_atts = fn_node.attributeCount()
    node_type_name = fn_node.typeName
    for a in range(num_atts):
        att_obj = fn_node.attribute(a)
        if att_obj.hasFn(om.MFn.kTypedAttribute):
            fn_attr = om.MFnTypedAttribute(att_obj)
            attr_type = fn_attr.attrType()
            attr_name = fn_attr.name

            if "{}.{}".format(node_type_name, attr_name) not in ATTR_BLACKLIST:
                if fn_attr.usedAsFilename:
                    result.append(attr_name)
                elif (
                    attr_type == om.MFnData.kString or attr_type == om.MFnData.kStringArray
                ):
                    attr_name_lower = attr_name.lower()
                    if "file" in attr_name_lower or "path" in attr_name_lower:
                        result.append(attr_name)

    return result


def get_filename_paths_for_node(fn_node, attnames):
    """
    Get filename values for usedAsFilename plugs.

    When we get a node's plug from the att name, it may be a child of a compound
    array plug (or even nested seveal levels). In this case the name will be of
    the form node_name.parentPlug[-1].childPlug This (-1) is a nonexistent plug.
    If any plugs do exist the index wil be >= 0. In order to get the actual plug
    elements, if they exist, we use a SelectionList. We add a wildcard name to
    the selectionlist and iterate through the real plugs it contains.

    The [-1] indicator is only used for array parent plugs, not leaf level array
    plugs.

    If the leaf level plug is an array, then we find out with isArray() and
    iterate over its elements.
    """
    ws = pm.Workspace()
    result = []
    for attname in attnames:
        plug = fn_node.findPlug(attname, False)
        plug_selection_list = om.MSelectionList()
        try:
            plug_selection_list.add(plug.name().replace("[-1]", "[*]"))
        except RuntimeError:
            continue

        num_plugs = plug_selection_list.length()
        for p in xrange(num_plugs):
            real_plug = plug_selection_list.getPlug(p)
            if real_plug.isArray:
                for physical_index in xrange(real_plug.numElements()):
                    child_plug = plug.elementByPhysicalIndex(physical_index)
                    value = child_plug.asString()
                    if value:
                        result.append(
                            {"path": ws.expandName(
                                value), "plug": child_plug.name()}
                        )
            else:
                print real_plug.name()
                value = real_plug.asString()
                if value:
                    result.append(
                        {"path": ws.expandName(
                            value), "plug": real_plug.name()}
                    )

    return result


def extend_with_tx_paths(paths):
    """
    Add the tx equivalent of every File Texture

    Use glob notation for the extension (.[t][x]) because when paths are finally
    resolved, the list is expanded by globbing. As TX files are not critical, we
    don't want to block the submission if they don't exist. Glob will expand to
    no files if the file does not exist.
    """
    txs = []
    for p in paths:
        if p["plug"].split(".")[-1] == "computedFileTextureNamePattern":
            root, _ = os.path.splitext(p["path"])
            txpath = "{}.[t][x]".format(root)
            txs.append({"path": txpath, "plug": p["plug"]})
    return paths+txs



"""
Handle the UI for 2 attributes:

1. hostSoftware (single value maya version)
2. pluginSoftware (multi value plugin name and version)

TODO: Try to move ALL the plugin/host compatibility logic out of this file.
"""

import pymel.core as pm
from ciocore import data as coredata
from ciomaya.lib import software
from ciomaya.lib import const as k
from ciomaya.lib.ae import AEcommon


def create_ui(node_attr):
    """Build static UI.

    A rowLayout for the host.
    A columnLayout for the plugins.
    """
    with AEcommon.ae_template():

        label = pm.text("hostLabel", label="Maya Version",
                        width=k.AE_TEXT_WIDTH)
        menu = pm.optionMenu("hostMenu", acc=True)
        button = pm.button("addButton", label="Add Plugin Entry", width=90)

        _form_layout(
            label,
            menu,
            button,
            pm.text(label="", width=30)
        )

        popup = pm.popupMenu(parent=label)
        pm.setParent(popup, menu=True)
        pm.menuItem(label="Detect Software")
        pm.menuItem(label="Add Plugin Entry")

        pm.setParent("..")  # out of form

        pm.columnLayout("pluginLayout", adj=True)

        pm.setParent("..")  # out of columnLayout

        populate_ui(node_attr)


def populate_ui(node_attr):
    """Reconfigure action buttons when node changes."""
    host_attr = pm.Attribute(node_attr)
    plugin_attr = host_attr.node().attr("pluginSoftware")
    widgets = _get_widgets()

    if not coredata.valid():
        for item in pm.optionMenu(widgets["host_menu"], q=True, itemListLong=True):
            pm.deleteUI(item)
        pm.setParent(widgets["host_menu"], menu=True)
        pm.menuItem(label="Not connected")

    # update popup menu items
    _configure_popup_menu(host_attr, plugin_attr, widgets)

    pm.optionMenu(widgets["host_menu"], edit=True, changeCommand=pm.Callback(
        _set_host_software_value, host_attr, plugin_attr, widgets))

    pm.button(widgets["add_button"], edit=True, command=pm.Callback(
        _add_plugin,  plugin_attr, "unknown"))

    # Update this UI if the attribute changes by some other means
    # For example: setAttr, or another instance of the attribute editor.
    _setup_script_jobs(host_attr, plugin_attr, widgets)

    pm.evalDeferred(pm.Callback(_ensure_connection,
                                host_attr, plugin_attr, widgets))


def _setup_script_jobs(host_attr, plugin_attr, widgets):

    menu = widgets["host_menu"]
    pm.scriptJob(
        attributeChange=(
            host_attr,
            pm.Callback(_sync_all_menus_to_attributes,
                        host_attr, plugin_attr,  widgets),
        ),
        parent=menu,
        replacePrevious=True
    )

    pm.scriptJob(
        attributeChange=(
            plugin_attr,
            pm.Callback(_sync_all_menus_to_attributes,
                        host_attr, plugin_attr,  widgets),
        ),
        parent=menu
    )


def _get_widgets(parent=None):
    if not parent:
        parent = pm.setParent(q=True)

    host_label = AEcommon.find_ui("hostLabel", parent)
    return {
        "host_layout": AEcommon.find_ui("hostLayout", parent),
        "plugin_layout": AEcommon.find_ui("pluginLayout", parent),
        "host_label": host_label,
        "host_menu": AEcommon.find_ui("hostMenu", parent),
        "add_button": AEcommon.find_ui("addButton", parent),
        "popup_menu":  pm.control(host_label, q=True, popupMenuArray=True)[0]
    }


def _ensure_connection(host_attr, plugin_attr, widgets):
    """Fetch list of sw from Conductor.

    Use cached data on the module if force is False"""

    software_data = coredata.data().get("software")
    if not coredata.valid():
        return

    AEcommon.ensure_populate_menu(
        widgets["host_menu"], software_data.supported_host_names())
    _sync_all_menus_to_attributes(host_attr, plugin_attr,  widgets)


def _configure_popup_menu(host_attr, plugin_attr, widgets):

    detect_item, add_plugin_item = pm.popupMenu(
        widgets["popup_menu"], q=True, itemArray=True)

    pm.menuItem(detect_item, edit=True, command=pm.Callback(
        _detect_software, host_attr, plugin_attr, widgets))

    pm.menuItem(add_plugin_item, edit=True, command=pm.Callback(
        _add_plugin,  plugin_attr, "unknown"))


def _detect_software(host_attr, plugin_attr, widgets):
    host_software_name = software.detect_host()
    host_attr.set(host_software_name)
    _remove_all_plugins(plugin_attr)
    for plugin_software_name in filter(None, [
        software.detect_mtoa(),
        software.detect_rfm(),
        software.detect_vray(),
        software.detect_redshift(),
        software.detect_yeti(),

    ]):
        _add_plugin(plugin_attr, plugin_software_name)


def _add_plugin(plugin_attr, plugin_name):
    AEcommon.next_available_element_plug(plugin_attr).set(plugin_name)


def _remove_all_plugins(plugin_attr):
    for index in plugin_attr.getArrayIndices():
        pm.removeMultiInstance(plugin_attr[index], b=True)


def _sync_all_menus_to_attributes(host_attr, plugin_attr,  widgets):

    AEcommon.sync_menu_to_attr(widgets["host_menu"], host_attr, default=-1)

    _sync_plugin_software_menus(
        host_attr, plugin_attr, widgets["plugin_layout"])


def _set_host_software_value(host_attr, plugin_attr, widgets):
    """
    Respond to host menu change.
    """

    num_items = pm.optionMenu(widgets["host_menu"], q=True, numberOfItems=True)
    if not num_items:
        return
    selected_value = pm.optionMenu(widgets["host_menu"], q=True, value=True)
    host_attr.set(selected_value)
    AEcommon.print_setAttr_cmd(host_attr)

    _sync_plugin_software_menus(
        host_attr, plugin_attr, widgets["plugin_layout"])


def _sync_plugin_software_menus(host_attr, plugin_attr, plugin_layout):
    """
     Make sure plugin menu items reflect the pluginSoftware attribute values.

     We simply delete all UI and rebuild from scratch.
    """

    for widget in (
        pm.layout(plugin_layout, q=True, childArray=True) or []
    ):
        pm.deleteUI(widget)

    if not plugin_attr.getArrayIndices():
        return

    software_data = coredata.data().get("software")
    supported_plugins = software_data.supported_plugins(host_attr.get())
    if not supported_plugins:
        return

    for element_attr in plugin_attr:
        if element_attr.get():
            pm.setParent(plugin_layout)

            _build_plugin_option_menus(element_attr, supported_plugins)


def _build_plugin_option_menus(attr, supported_plugins):
    """
    Build 2 menus (plugin, version) for the one plugin at index.

    The plugin at node.attr("pluginSoftware")[index] is something like:
    "vray 1.2.3" so we check that it is supported by the current maya
    version, and if not we reset the attr to one that is. Then we build the
    two menus.
    1. plugin menu contains all plugin names supported by current maya
        version.
    2. version menu contains versions of the currently selected plugin.
    """

    valid_plugin = _ensure_valid_selected_plugin(attr, supported_plugins)
    if not valid_plugin:
        return

    plugin, version = valid_plugin

    label = pm.text(label="Plugin Version", width=k.AE_TEXT_WIDTH)
    plugin_menu = pm.optionMenu(acc=True)
    version_menu = pm.optionMenu(acc=True)
    remove_btn = pm.symbolButton(
        image="smallTrash.xpm", ann="Remove plugin")

    form = _form_layout(
        label,
        plugin_menu,
        version_menu,
        remove_btn
    )

    pm.setParent(plugin_menu, menu=True)
    for p in supported_plugins:
        pm.menuItem(label=p["plugin"])

    pm.optionMenu(
        plugin_menu,
        edit=True,
        changeCommand=pm.Callback(
            _plugin_menu_change, attr, supported_plugins, plugin_menu, version_menu)
    )

    pm.optionMenu(
        version_menu,
        edit=True,
        changeCommand=pm.Callback(
            _version_menu_change, attr, plugin_menu, version_menu
        ),
    )
    pm.symbolButton(remove_btn, edit=True,
                    command=pm.Callback(_on_remove_plugin, attr))

    plugin_index = _get_plugin_index(plugin, supported_plugins)
    versions = _get_versions_for_supported_plugin(plugin, supported_plugins)
    pm.optionMenu(plugin_menu, edit=True, sl=(plugin_index+1))

    # now buildversion menu
    pm.setParent(version_menu, menu=True)
    for v in versions:
        pm.menuItem(label=v)

    version_index = _get_version_index(version, versions)
    pm.optionMenu(version_menu, edit=True, sl=(version_index+1))


def _ensure_valid_selected_plugin(attr, supported_plugins):
    if not supported_plugins:
        return
    current = attr.get()
    try:
        current_plugin, current_version = current.split(" ")
        found_plugin = next(
            (x for x in supported_plugins if x["plugin"] == current_plugin), None)
        if found_plugin:
            found_version = next(
                (x for x in found_plugin["versions"]
                 if x == current_version), None
            )
            if not found_version:
                current_version = found_plugin["versions"][-1]
        else:
            current_plugin = supported_plugins[0]["plugin"]
            current_version = supported_plugins[0]["versions"][-1]
    except (ValueError, AttributeError):
        current_plugin = supported_plugins[0]["plugin"]
        current_version = supported_plugins[0]["versions"][-1]
    current_plugin_value = "{} {}".format(current_plugin, current_version)
    if attr.get() != current_plugin_value:
        attr.set(current_plugin_value)
        AEcommon.print_setAttr_cmd(attr)

    return (current_plugin, current_version)


def _plugin_menu_change(attr, supported_plugins, plugin_menu, version_menu):
    """Callback when plugin menu item is selected.

    We rebuild the versions menu for this plugin, and select the last
    element.
    """
    plugin = pm.optionMenu(plugin_menu, q=True, value=True)
    versions = _get_versions_for_supported_plugin(plugin, supported_plugins)

    pm.setParent(version_menu, menu=True)
    for item in pm.optionMenu(version_menu, q=True, itemListLong=True):
        pm.deleteUI(item)
    for v in versions:
        pm.menuItem(label=v)
    pm.optionMenu(version_menu, edit=True, sl=len(versions))

    attr.set("{} {}".format(plugin, versions[-1]))
    AEcommon.print_setAttr_cmd(attr)


def _version_menu_change(attr, plugin_menu, version_menu):
    """Callback when version menu item is selected.

    Set the attribute to the string combination of 'plugin version'
    """
    plugin = pm.optionMenu(plugin_menu, q=True, value=True)
    version = pm.optionMenu(version_menu, q=True, value=True)
    value = "{} {}".format(plugin, version)
    if value != attr.get():
        attr.set(value)
        AEcommon.print_setAttr_cmd(attr)


def _on_remove_plugin(attr):
    """Remove attr multi element.
    Redraw is triggered automatically.
    """
    pm.removeMultiInstance(attr, b=True)


def _get_versions_for_supported_plugin(plugin, plugins):
    return next((x["versions"] for x in plugins if x["plugin"] == plugin), None)


def _get_plugin_index(plugin, plugins):
    return next((i for i, x in enumerate(plugins) if x["plugin"] == plugin), None)


def _get_version_index(version, versions):
    return next((i for i, x in enumerate(versions) if x == version), None)


def _form_layout(*widgets):
    # There must be 4 widgets.
    # we stretch the second and third widget.
    form = pm.formLayout(nd=100)
    for widget in widgets:
        pm.control(widget, edit=True, parent=form)

    form.attachForm(widgets[0], "left", 2)
    form.attachNone(widgets[0], "right")
    form.attachForm(widgets[0], "top", 2)
    form.attachForm(widgets[0], "bottom", 2)

    form.attachControl(widgets[1], "left", 2, widgets[0])
    form.attachPosition(widgets[1], "right", 2,  65)
    form.attachForm(widgets[1], "top", 2)
    form.attachForm(widgets[1], "bottom", 2)

    form.attachControl(widgets[2], "left", 2, widgets[1])
    form.attachControl(widgets[2], "right", 2, widgets[3])
    form.attachForm(widgets[2], "top", 2)
    form.attachForm(widgets[2], "bottom", 2)

    form.attachNone(widgets[3], "left")
    form.attachForm(widgets[3], "right", 2)
    form.attachForm(widgets[3], "top", 2)
    form.attachForm(widgets[3], "bottom", 2)

    return form



"""
Submit.

"""
import json
import os
import sys
import traceback
from contextlib import contextmanager

import pymel.core as pm
from ciocore import conductor_submit
from ciocore.expander import Expander
from ciomaya.lib import const as k
from ciomaya.lib import layer_utils, validation, window
from ciocore.validator  import ValidationError
 

@contextmanager
def full_output(node):
    task_limit = node.attr("taskLimit").get()
    do_scrape = node.attr("doScrape").get()
    node.attr("taskLimit").set(-1)
    node.attr("doScrape").set(True)
    yield
    node.attr("taskLimit").set(task_limit)
    node.attr("doScrape").set(do_scrape)


@contextmanager
def transient_save(filepath, cleanup=True):
    original = pm.sceneName()
    pm.saveAs(filepath)
    yield
    pm.renameFile(original)
    if cleanup:
        try:
            os.remove(filepath)
        except OSError:
            pm.displayWarning("Couldn't cleanup file: {}".format(filepath))


def valid(node):

    try:
        validation.run(node)
    except ValidationError as ex:
        pm.displayWarning(str(ex))
        return False
    return True


def submit(node):
    # Dependencies are cached the first time calculated so that we dont have to
    # scrape for every layer.
    filepath = pm.sceneName()
    if filepath and node.attr("autosave").get():
        cleanup = node.attr("cleanupAutosave").get(
        ) and not node.attr("useUploadDaemon").get()
        filepath = _resolve_autosave_template(node)
        with transient_save(filepath, cleanup=cleanup):
            if valid(node):
                handle_submissions(node)
        return

    if pm.isModified():
        filepath = browse_save_as()
        if not filepath:
            pm.warning('No file Selected')
            return
        pm.saveAs(filepath)

    if valid(node):
        handle_submissions(node)


def handle_submissions(node):
    submissions = get_submissions(node)
    responses = do_submissions(submissions, node)
    window.show_submission_responses(responses)


def get_submissions(node):

    submissions = []
    layer_policy = node.attr("renderLayers").get()

    if layer_policy == k.CURRENT_LAYER:
        submissions.append(get_submission(node))
    else:
        for layer in layer_utils.get_renderable_legacy_layers():
            with layer_utils.layer_context(layer):
                submissions.append(get_submission(node))
    return list(filter(None, submissions))


def do_submissions(submissions, node):
    results = []
    show_tracebacks = node.attr("showTracebacks").get()
    for submission in submissions:
        try:
            remote_job = conductor_submit.Submit(submission)
            response, response_code = remote_job.main()
            results.append({"code": response_code, "response": response})
        except BaseException as ex:
            if show_tracebacks:
                msg = traceback.format_exc()
            else:
                msg = ex.message
            pm.displayError(msg)
            results.append({"code": "undefined", "response": msg})

    return results


def _resolve_autosave_template(node):

    scene_path = pm.sceneName()
    if not scene_path:
        scene_path = os.path.join(
            pm.workspace.getPath(),
            pm.workspace.fileRules.get("mayaAscii").split(":")[0],
            "untitled.ma"
        )

    scene_name = os.path.splitext(os.path.split(scene_path)[1])[0]
    context = {
        "Scene": scene_name,
        "scene": scene_name
    }
    expander = Expander(**context)
    resolved_name = expander.evaluate(node.attr("autosaveTemplate").get())
    return os.path.join(os.path.dirname(scene_path), "{}.ma".format(resolved_name))


def browse_save_as():
    filters = "Maya Files (*.ma *.mb);;Maya ASCII (*.ma);;Maya Binary (*.mb);;All Files (*.*)"

    entries = pm.fileDialog2(
        caption="Save File As",
        okCaption="Save As",
        fileFilter=filters,
        dialogStyle=2,
        fileMode=0,
        dir=os.path.dirname(pm.sceneName()))

    if entries:
        return entries[0]


def get_submission(node):
    out_attr = pm.PyNode(node).attr("output")
    with full_output(node):
        pm.dgdirty(out_attr)
        result = out_attr.get()
        if result:
            return json.loads(result)



"""
Handle the UI for projects:

"""

import pymel.core as pm
from ciocore import data as coredata
from ciomaya.lib import const as k
from ciomaya.lib.ae import AEcommon


def create_ui(node_attr):
    """Build static UI"""

    with AEcommon.ae_template():

        pm.rowLayout(
            numberOfColumns=2,
            columnWidth2=(k.AE_TEXT_WIDTH, 300),
            columnAttach=((1, "right", 0), (2, "both", 0)),
        )

        pm.text("projectMenuLabel", label="Conductor project")
        pm.optionMenu("projectMenu", acc=True)


        pm.setParent("..")

        populate_ui(node_attr)


def populate_ui(node_attr):
    """Populate / reconfigure UI for the current node"""
    attr = pm.Attribute(node_attr)
    widgets = _get_widgets()

    if not coredata.valid():
        for item in pm.optionMenu(widgets["menu"], q=True, itemListLong=True):
            pm.deleteUI(item)
        pm.setParent(widgets["menu"], menu=True)
        pm.menuItem(label="Not connected")

    pm.optionMenu(widgets["menu"], edit=True, changeCommand=pm.Callback(
        _set_project_value, attr, widgets["menu"]))

    # Update this UI if the attribute changes by some other means
    # For example: setAttr, or another instance of the attribute editor.
    _setup_script_jobs(attr, widgets)

    pm.evalDeferred(pm.Callback(_ensure_connection, attr, widgets))


def _get_widgets(parent=None):
    if not parent:
        parent = pm.setParent(q=True)
    return {
        "menu": AEcommon.find_ui("projectMenu", parent)
    }


def _setup_script_jobs(attr, widgets):
    menu = widgets["menu"]

    pm.scriptJob(
        attributeChange=(
            attr,
            pm.Callback(AEcommon.sync_menu_to_attr, menu, attr),
        ),
        parent=menu,
        replacePrevious=True
    )


def _ensure_connection(attr, widgets):
    """Fetch a fresh list of projects from Conductor.

    Use cached data on the module if force is False"""
    project_data = coredata.data()["projects"]
    AEcommon.ensure_populate_menu(widgets["menu"], project_data)

    AEcommon.sync_menu_to_attr(widgets["menu"], attr)


def _set_project_value(attr, menu):
    """
    Respond to menu change.
    """
    num_items = pm.optionMenu(menu, q=True, numberOfItems=True)
    if not num_items:
        return
    selected_value = pm.optionMenu(menu, q=True, value=True)
    if attr.get() != selected_value:
        attr.set(selected_value)
        AEcommon.print_setAttr_cmd(attr)



"""
Handle the UI for environment:
"""

import pymel.core as pm
from ciomaya.lib import const as k
from ciomaya.lib.ae import AEcommon


def create_ui(node_attr):
    """Build static UI."""
    with AEcommon.ae_template():

        pm.rowLayout(
            width=290,
            numberOfColumns=2,
            columnWidth2=(k.AE_TEXT_WIDTH, 210),
            columnAttach=((1, "right", 0), (2, "both", 0)),
        )

        pm.text(label="")
        pm.button("environmentAddButton",
                  label="Add Environment Variable", height=24)
        pm.setParent("..")  # out rowLayout
        pm.separator()

        _form_layout(pm.text(label="Key", align="left", width=k.AE_TEXT_WIDTH),
                     pm.text(label="Value", align="left"),
                     pm.text(label="Excl", align="left", width=36),
                     pm.text(label="", width=36)
                     )

        pm.setParent("..")  # out _form_layout

        pm.columnLayout("environmentColumn", adj=True)

        pm.setParent("..")  # out columnLayout

        populate_ui(node_attr)


def populate_ui(node_attr):
    """Reconfigure action buttons when node changes."""

    attr = pm.Attribute(node_attr)
    widgets = _get_widgets()

    pm.button(widgets["button"], edit=True,
              command=pm.Callback(_add_entry, attr))

    for widget in pm.columnLayout(widgets["column"], q=True, childArray=True) or []:
        pm.deleteUI(widget)

    pm.setParent(widgets["column"])
    for attr_element in attr:

        key_att = attr_element.attr("extraEnvironmentKey")
        val_att = attr_element.attr("extraEnvironmentValue")
        exc_att = attr_element.attr("extraEnvironmentExclusive")

        key_ctl = pm.textField(
            placeholderText="VAR_NAME", width=k.AE_TEXT_WIDTH)
        val_ctl = pm.textField(placeholderText="Value")
        exc_ctl = pm.checkBox(label="", width=36)
        del_ctl = pm.symbolButton(
            image="smallTrash.xpm", width=36, command=pm.Callback(
                pm.removeMultiInstance, attr_element, b=True
            )
        )

        _form_layout(key_ctl, val_ctl, exc_ctl, del_ctl)
        pm.setParent("..")  # out _form_layout

        pm.connectControl(key_ctl, key_att)
        pm.connectControl(val_ctl, val_att)
        pm.connectControl(exc_ctl, exc_att)

        pm.setParent(widgets["column"])


def _get_widgets(parent=None):
    if not parent:
        parent = pm.setParent(q=True)
    return {
        "button":  AEcommon.find_ui("environmentAddButton", parent),
        "column":  AEcommon.find_ui("environmentColumn", parent)
    }


def _add_entry(attr):
    AEcommon.next_available_element_plug(
        attr).attr("extraEnvironmentKey").set("")


def _form_layout(*widgets):
    # There must be 3 widgets.
    form = pm.formLayout(nd=100)
    for widget in widgets:
        pm.control(widget, edit=True, parent=form)

    form.attachForm(widgets[0], "left", 2)
    form.attachNone(widgets[0], "right")
    form.attachForm(widgets[0], "top", 2)
    form.attachForm(widgets[0], "bottom", 2)

    form.attachControl(widgets[1], "left", 2, widgets[0])
    form.attachControl(widgets[1], "right", 2, widgets[2])
    form.attachForm(widgets[1], "top", 2)
    form.attachForm(widgets[1], "bottom", 2)

    form.attachNone(widgets[2], "left")
    form.attachControl(widgets[2], "right", 2, widgets[3])
    form.attachForm(widgets[2], "top", 2)
    form.attachForm(widgets[2], "bottom", 2)

    form.attachNone(widgets[3], "left")
    form.attachForm(widgets[3], "right", 2)
    form.attachForm(widgets[3], "top", 2)
    form.attachForm(widgets[3], "bottom", 2)

    return form


class AEenvironment(object):
    def __init__(self, aet):
        self.add_btn = None
        self.col = None
        aet.callCustom(self.create_ui, self.populate_ui, "extraEnvironment")

    def create_ui(self, node_attr):
        """Build static UI"""
        pm.setUITemplate("attributeEditorTemplate", pushTemplate=True)
        node = pm.Attribute(node_attr).node()
        pm.rowLayout(
            width=290,
            numberOfColumns=2,
            columnWidth2=(k.AE_TEXT_WIDTH, 210),
            columnAttach=((1, "right", 0), (2, "both", 0)),
        )

        pm.text(label="")
        self.add_btn = pm.button(
            label="Add Environment Variable",
            height=24,
            command=pm.Callback(on_add_btn, node),
        )
        pm.setParent("..")
        pm.separator()
        pm.rowLayout(
            width=290,
            numberOfColumns=4,
            adjustableColumn=2,
            columnWidth4=(k.AE_TEXT_WIDTH, 150, 36, 24),
            columnAttach=(
                (1, "both", 6),
                (2, "both", 6),
                (3, "both", 6),
                (4, "both", 0),
            ),
        )
        pm.text(label="Key", align="left")
        pm.text(label="Value", align="left")
        pm.text(label="Excl", align="left")
        pm.text(label="")
        pm.setParent("..")

        self.col = pm.columnLayout(adj=True)
        pm.setUITemplate(ppt=True)
        self.populate_ui(node_attr)

    def populate_ui(self, node_attr):
        """Reconfigure UI for the current node"""
        pm.setUITemplate("attributeEditorTemplate", pushTemplate=True)
        node = pm.Attribute(node_attr).node()
        pm.button(self.add_btn, edit=True,
                  command=pm.Callback(on_add_btn, node))

        for widget in pm.columnLayout(self.col, q=True, childArray=True) or []:
            pm.deleteUI(widget)

        pm.setParent(self.col)
        for attr in pm.Attribute(node_attr):
            pm.rowLayout(
                width=290,
                numberOfColumns=4,
                adjustableColumn=2,
                columnWidth4=(k.AE_TEXT_WIDTH, 150, 36, 24),
                columnAttach=(
                    (1, "both", 0),
                    (2, "both", 0),
                    (3, "both", 0),
                    (4, "both", 0),
                ),
            )

            key_att = attr.attr("extraEnvironmentKey")
            val_att = attr.attr("extraEnvironmentValue")
            exc_att = attr.attr("extraEnvironmentExclusive")

            key_tf = pm.textField(text=key_att.get(),
                                  placeholderText="VAR_NAME")
            val_tf = pm.textField(text=val_att.get(), placeholderText="Value")
            exc_cb = pm.checkBox(value=exc_att.get(), label="")
            pm.symbolButton(
                image="smallTrash.xpm", command=pm.Callback(on_remove_var, attr)
            )

            pm.textField(
                key_tf,
                edit=True,
                changeCommand=pm.Callback(on_text_changed, key_tf, key_att),
            )
            pm.textField(
                val_tf,
                edit=True,
                changeCommand=pm.Callback(on_text_changed, val_tf, val_att),
            )
            pm.checkBox(
                exc_cb,
                edit=True,
                changeCommand=pm.Callback(on_excl_cb_changed, exc_cb, exc_att),
            )

            pm.setParent(self.col)

        pm.setUITemplate(ppt=True)


def on_add_btn(node):
    indices = node.attr("extraEnvironment").getArrayIndices()
    next_available = next(a for a, b in enumerate(indices + [-1]) if a != b)
    node.attr("extraEnvironment")[next_available].attr(
        "extraEnvironmentKey").set("")
    node.attr("extraEnvironment")[next_available].attr(
        "extraEnvironmentValue").set("")


def on_text_changed(text_field, attribute):
    val = pm.textField(text_field, q=True, text=True)
    attribute.set(val)


def on_excl_cb_changed(checkbox, attribute):
    val = pm.checkBox(checkbox, q=True, value=True)
    attribute.set(val)


def on_remove_var(attribute):
    pm.removeMultiInstance(attribute, b=True)



"""
Handle the UI for extra assets:
"""

import pymel.core as pm
from ciomaya.lib import const as k

from ciomaya.lib.ae import AEcommon


def create_ui(node_attr):
    """Build static UI.
    """
    attr = pm.Attribute(node_attr)
    with AEcommon.ae_template():

        form = pm.formLayout("ddFormName", nd=100)

        field = pm.textFieldGrp("destinationField", label="Destination Directory")

        button = pm.symbolButton(
            "destinationButton",
            image="SP_DirClosedIcon.png", width=24, height=24
        )
        pm.setParent("..")  # out of formLayout

        form.attachNone(button, "left")
        form.attachForm(button, "right", 2)
        form.attachForm(button, "top", 2)
        form.attachForm(button, "bottom", 2)

        form.attachForm(field, "left", 2)
        form.attachControl(field, "right", 2, button)
        form.attachForm(field, "top", 2)
        form.attachForm(field, "bottom", 2)

        populate_ui(node_attr)


def populate_ui(node_attr):
    """Reconfigure action buttons when node changes."""
    widgets = _get_widgets()

    attr = pm.Attribute(node_attr)
    val =attr.get()
    pm.textFieldGrp(
         widgets["field"], 
         edit=True, 
         text=val, 
        changeCommand=pm.Callback(on_text_changed,  widgets["field"], attr)
     )

    pm.symbolButton(
        widgets["button"],
        edit=True,
        command=pm.Callback(_on_browse_button, attr,  widgets["field"]),
    )

def _get_widgets(parent=None):
    if not parent:
        parent = pm.setParent(q=True)
    return {
        "field":  AEcommon.find_ui("destinationField", parent),
        "button": AEcommon.find_ui("destinationButton", parent)
    }


def _on_browse_button(attr, field):
    path = browse_for_dest_directory()
    if path:
        attr.set(path)
        pm.textFieldGrp(field, edit=True, text=path)
        AEcommon.print_setAttr_cmd(attr)


def browse_for_dest_directory():
    entries = pm.fileDialog2(
        caption="Choose Directory",
        okCaption="Choose",
        fileFilter="*",
        dialogStyle=2,
        fileMode=3,
        dir=pm.workspace.getPath(),
    )
    if entries:
        return entries[0]
    pm.displayWarning("No files Selected")


def on_text_changed(text_field, attribute):
    val = pm.textFieldGrp(text_field, q=True, text=True)
    attribute.set(val)



"""
Handle the UI for instanceTypes:

"""

import maya.app.renderSetup.model.renderSetup as rs
import maya.app.renderSetup.views.overrideUtils as ov_utils
import pymel.core as pm
from ciocore import data as coredata
from ciomaya.lib import const as k
from ciomaya.lib.ae import AEcommon
 
def create_ui(node_attr):
    """Build static UI"""
    with AEcommon.ae_template():

        pm.rowLayout(
            numberOfColumns=2,
            columnWidth2=(k.AE_TEXT_WIDTH, 300),
            columnAttach=((1, "right", 0), (2, "both", 0)),
        )

        label = pm.text("instanceTypesLabel",label="Instance Type Name")
        pm.optionMenu("instanceTypesMenu", acc=True)
        popup = pm.popupMenu(parent=label)
        pm.setParent(popup, menu=True)
        pm.menuItem(label="Create Absolute Override for Visible Layer")

        pm.setParent("..")  # out of rowLayout

        populate_ui(node_attr)


def populate_ui(node_attr):
    """Reconfigure UI for the current node"""
    attr = pm.Attribute(node_attr)
    widgets = _get_widgets()
 
    if not  coredata.valid():
        for item in pm.optionMenu(widgets["menu"], q=True, itemListLong=True):
            pm.deleteUI(item)
        pm.setParent(widgets["menu"], menu=True)
        pm.menuItem(label="Not connected")

    # update popup menu items
    _configure_popup_menu(attr, widgets)

    pm.optionMenu(widgets["menu"], edit=True, changeCommand=pm.Callback(
        _set_instance_type_value, attr, widgets["menu"]))
    # Update this UI if the attribute changes by some other means
    # For example: setAttr, or another instance of the attribute editor.
    _setup_script_jobs(attr, widgets)

    _set_label_color(attr, widgets["label"])

    pm.evalDeferred(pm.Callback(_ensure_connection, attr, widgets))


def _configure_popup_menu(attr, widgets):
    override_item = pm.popupMenu(
        widgets["popup_menu"], q=True, itemArray=True)[0]

    enable_override = (
        pm.editRenderLayerGlobals(query=True, currentRenderLayer=True)
        != "defaultRenderLayer"
    )
    pm.menuItem(override_item, edit=True, en=enable_override, 
      command=pm.Callback(_create_layer_override, attr, widgets["label"]) )
 

def _get_widgets(parent=None):
    if not parent:
        parent = pm.setParent(q=True)
    label = AEcommon.find_ui("instanceTypesLabel", parent),
    return {
        "label" :label,
        "menu": AEcommon.find_ui("instanceTypesMenu", parent),
        "popup_menu": pm.control( label, q=True, popupMenuArray=True)[0]
    }



def _setup_script_jobs(attr, widgets):
    menu = widgets["menu"]

    pm.scriptJob(
        attributeChange=(
            attr,
            pm.Callback(_sync_menu_to_attr, attr, widgets),
        ),
        parent=menu,
        replacePrevious=True
    )

    pm.scriptJob(
        event=(
            "renderLayerManagerChange",
            pm.Callback(_on_render_layer_manager_change, attr, widgets),
        ),
        parent=menu,
    )


def _on_render_layer_manager_change(attr, widgets):

    _sync_menu_to_attr(attr, widgets)
    _set_label_color(attr, widgets["label"])


def _ensure_connection(attr, widgets, force=False):
    """Fetch a fresh list of inst types from Conductor (or the cache)."""

    instance_type_data = coredata.data(force_instance_types=force)["instance_types"]
    descriptions =  [t["description"] for t in instance_type_data]
    if len(descriptions) and isinstance(descriptions[0], list):
        descriptions = [v[0] for v in descriptions]

    AEcommon.ensure_populate_menu(widgets["menu"], descriptions)

    values =  [it["name"] for it in instance_type_data]
    AEcommon.sync_menu_to_attr(widgets["menu"], attr , values=values)


def _sync_menu_to_attr(attr, widgets):
    """
    Make sure menu item reflects the attribute value.

    If the attribute is invalid, set it to the first valid project.


    """
    menu = widgets["menu"]
    label = widgets["label"]
    
    items = pm.optionMenu(menu, q=True, itemListLong=True)
    if not items:
        return

    name = attr.get()
    index, instance_type = find_by_key("name", name)
    if not instance_type:
        pm.displayWarning("Didn't find '{}' in instance types".format(name))
        return

    one_based_index = index+1
    if one_based_index > len(items):
        one_based_index = 1
        attr.set(coredata.data()["instance_types"][0]["name"])

    pm.optionMenu(menu, edit=True, sl=one_based_index)

    _set_label_color(attr, label)


def _set_instance_type_value(attr, menu):
    """
    Respond to menu change.
    """
    num_items = pm.optionMenu(menu, q=True, numberOfItems=True)
    if not num_items:
        return
    description = pm.optionMenu(menu, q=True, value=True)
    _, instance_type = find_by_key("description", description)
    if not instance_type:
        pm.displayWarning("Didn't find '{}' in instance types".format(description))
        return

    name = instance_type["name"]
    if attr.get() != name:
        attr.set(name)
        AEcommon.print_setAttr_cmd(attr)


def _create_layer_override(attr, label):
    ov_utils.createAbsoluteOverride( attr.node().name(), attr.attrName(True) )
    _set_label_color(attr, label)



def _set_label_color(attr, label):
    """By convention, label is orange if attr has an override."""
    has_override = rs.hasOverrideApplied(
        attr.node().name(),  attr.attrName(True))
    text = "Instance Type Name"
    label_text = "<font color=#ec6a17>{}</font>".format(
        text) if has_override else text
    pm.text(label, edit=True, label=label_text)



def find_by_key(key, value):
    result = (None,None)
    if not coredata.valid():
        return result
    try:
        result = next(tup for tup in enumerate(coredata.data()["instance_types"]) if tup[1][key] == value)
    except StopIteration:
        pass
    return result
 




from ciocore.gpath_list import PathList
import importlib


__data__ = None


def data(node):
    global __data__

    if __data__:
        return __data__

    scrapers = []
    attr = node.attr("assetScrapers")
    for attr_element in attr:
        if not attr_element.attr("assetScraperActive").get():
            continue
        script = attr_element.attr("assetScraperName").get().strip()
        if not script:
            continue
        scrapers.append(script)

    result = []
    for script in scrapers:
        scraper_module = importlib.import_module(script)
        reload(scraper_module)
        paths = scraper_module.run(node)
        result += [p["path"] for p in paths if p["path"].strip()]

    result += list(node.attr("extraAssets").get())

    __data__ = PathList(*result)

    return __data__


def clear():
    global __data__
    __data__ = None



"""
Handle the UI for extra assets:
"""


import importlib

import os
import glob
import pymel.core as pm
from ciomaya.lib import const as k
from ciomaya.lib.ae import AEcommon
import traceback

def has_letters(rhs):
    return rhs.lower().islower()


def get_module_docstring(script):
    try:
        doctring = importlib.import_module(script).__doc__
        if has_letters(doctring):
            return doctring.strip()
    except BaseException:
        pass
    return "No doctring"


def create_ui(node_attr):
    """Build static UI."""
    with AEcommon.ae_template():

        f = pm.frameLayout(lv=False, cll=False, cl=False)
        pm.rowLayout("addButtonRow",
                     numberOfColumns=3,
                     columnWidth3=(k.AE_TEXT_WIDTH, 120, 120),
                     columnAttach=((1, 'both', 0),
                                   (2, 'both', 0),
                                   (3, 'both', 0)))

        pm.text(label="")
        pm.button("addButton",
                  label="Add Scraper", height=24
                  )
        pm.button("resetButton",
                  label="Reset Defaults", height=24
                  )
        pm.setParent("..")

        pm.separator()
        pm.columnLayout("scrapersColumn", adj=True)
        pm.setParent("..")
        pm.separator()

        btn = pm.symbolButton("testAllButton",
                              image="SP_FileIcon.png", ann="Test active scrapers", width=24, height=24)

        _form_layout(
            pm.text(label=""),
            pm.text(label=""),
            pm.text(label="Test All"),
            btn,
            pm.text(label="", width=25)
        )

        pm.setParent(f)
        pm.setParent("..")

        populate_ui(node_attr)


def populate_ui(node_attr):
    """Reconfigure action buttons when node changes."""

    attr = pm.Attribute(node_attr)
    widgets = _get_widgets()

    pm.button(widgets["add_button"], edit=True,
              command=pm.Callback(_on_add_btn, attr))

    pm.button(widgets["reset_button"], edit=True,
              command=pm.Callback(_on_reset_btn, attr))

    for widget in pm.columnLayout(widgets["scrapers_column"], q=True, childArray=True) or []:
        pm.deleteUI(widget)

    pm.setParent(widgets["scrapers_column"])
    for attr_element in attr:

        active_att = attr_element.attr("assetScraperActive")
        path_att = attr_element.attr("assetScraperName")

        active_ctl = pm.checkBox(label="")
        path_ctl = pm.textField()

        browse_ctl = pm.symbolButton(image="SP_DirClosedIcon.png", width=24, height=24,
                                     command=pm.Callback(_on_browse_button, attr_element, path_ctl, active_ctl))

        test_ctl = pm.symbolButton(image="SP_FileIcon.png", ann="Test this scraper", width=24, height=24,
                                   command=pm.Callback(_test_scraper, attr_element))

        del_ctl = pm.symbolButton(image="smallTrash.xpm", width=24, height=24,
                                  command=pm.Callback(_remove_scraper, attr_element))

        # arrange these controls in a formLayout. Easier than a rowLayout
        _form_layout(active_ctl, path_ctl, browse_ctl, test_ctl, del_ctl)

        pm.connectControl(active_ctl, active_att)
        pm.connectControl(path_ctl, path_att)

        _setup_script_jobs(attr_element, path_ctl)

        _active_cb_changed(active_att, path_ctl)
        _set_docstring(path_att, path_ctl)

        pm.setParent(widgets["scrapers_column"])

    pm.setParent("..")

    pm.symbolButton(widgets["test_all_button"], edit=True,  command=pm.Callback(
        _test_scraper, attr))


def _setup_script_jobs(attr_element, path_ctl):
    active_att = attr_element.attr("assetScraperActive")
    path_att = attr_element.attr("assetScraperName")

    pm.scriptJob(
        attributeChange=(
            active_att,
            pm.Callback(_active_cb_changed, active_att, path_ctl),
        ),
        parent=path_ctl,
        replacePrevious=True
    )

    pm.scriptJob(
        attributeChange=(
            path_att,
            pm.Callback(_set_docstring, path_att, path_ctl)
        ),
        parent=path_ctl
    )


def _form_layout(*widgets):
    # There must be 5 widgets.
    # we stretch the second widget.
    form = pm.formLayout(nd=100)
    for widget in widgets:
        pm.control(widget, edit=True, parent=form)

    form.attachForm(widgets[0], "left", 2)
    form.attachNone(widgets[0], "right")
    form.attachForm(widgets[0], "top", 2)
    form.attachForm(widgets[0], "bottom", 2)

    form.attachControl(widgets[1], "left", 2, widgets[0])
    form.attachControl(widgets[1], "right", 2, widgets[2])
    form.attachForm(widgets[1], "top", 2)
    form.attachForm(widgets[1], "bottom", 2)

    form.attachNone(widgets[2], "left")
    form.attachControl(widgets[2], "right", 2, widgets[3])
    form.attachForm(widgets[2], "top", 2)
    form.attachForm(widgets[2], "bottom", 2)

    form.attachNone(widgets[3], "left")
    form.attachControl(widgets[3], "right", 2, widgets[4])
    form.attachForm(widgets[3], "top", 2)
    form.attachForm(widgets[3], "bottom", 2)

    form.attachNone(widgets[4], "left")
    form.attachForm(widgets[4], "right", 2)
    form.attachForm(widgets[4], "top", 2)
    form.attachForm(widgets[4], "bottom", 2)

    return form


def _get_widgets(parent=None):
    if not parent:
        parent = pm.setParent(q=True)
    return {
        "add_button":  AEcommon.find_ui("addButton", parent),
        "reset_button":  AEcommon.find_ui("resetButton", parent),
        "scrapers_column":  AEcommon.find_ui("scrapersColumn", parent),
        "test_all_button":  AEcommon.find_ui("testAllButton", parent)
    }


def _on_add_btn(attr):
    script = _browse_for_script()
    if script:
        script = os.path.splitext(os.path.basename(script))[0]
        if script:
            AEcommon.next_available_element_plug(
                attr).attr("assetScraperName").set(script)


def _on_reset_btn(attr):
    for element in attr:
        pm.removeMultiInstance(element, b=True)

    script_path = os.path.join(
        pm.moduleInfo(
            path=True, moduleName="conductor"), "scripts"
    )
    files = sorted(glob.glob("{}/scrape_*.py".format(script_path)))
    for i, scraper in enumerate(files):
        attr[i].attr("assetScraperName").set(os.path.splitext(os.path.basename(scraper))[0])


def _on_browse_button(attr, path_field, active_checkbox):
    script = _browse_for_script()
    if script:
        script = os.path.splitext(os.path.basename(script))[0]
        attr.attr("assetScraperName").set(script)
        attr.attr("assetScraperActive").set(True)
        pm.textField(path_field, edit=True,   enable=True)
        pm.checkBox(active_checkbox, edit=True, value=True)
        pm.evalDeferred(pm.Callback(
            _set_docstring, attr.attr("assetScraperName"), path_field))


def _browse_for_script():
    entries = pm.fileDialog2(
        caption="Choose Script",
        okCaption="Choose",
        fileFilter="*",
        dialogStyle=2,
        fileMode=1,
        dir=pm.workspace.getPath())
    if entries:
        return entries[0]
    pm.displayWarning('No files Selected')


def _test_scraper(attr):
    node = attr.node()
    if attr.isArray():
        scripts = [attr_element.attr("assetScraperName").get(
        ) for attr_element in attr if attr_element.attr("assetScraperActive").get()]
    else:
        scripts = [attr.attr("assetScraperName").get()]
    scripts_repr = repr(scripts)
    paths = _run_scrapers(node, scripts)
    if paths:
        pm.displayInfo("Paths for scrapers: {}".format(scripts_repr))
        for path in paths:
            print path
    else:
        pm.displayInfo(
            "Scrapers returned no filenames: {}".format(scripts_repr))


def _run_scrapers(node, scripts):
    result = []
    show_tracebacks = node.attr("showTracebacks").get()
    msg = "Please open the Asset scraper section and test individual scrapers. Remove any offending scraper entries."
    for script in scripts:
        try:
            scraper_module = importlib.import_module(script)
            reload(scraper_module)
            result += scraper_module.run(node)
        except SyntaxError:
            pm.displayError(
                "Syntax error in the scraper script: '{}'.\n{}".format(script, msg))
            if show_tracebacks:
                 pm.displayError( traceback.format_exc())
            raise
        except ImportError:
            pm.displayError(
                "Can't load the script '{}' as a Python module.\n{}".format(script, msg))
            if show_tracebacks:
                 pm.displayError( traceback.format_exc())
            raise
        except BaseException:
            pm.displayError(
                "Unknown problem with the script '{}'.\n{}".format(script, msg))
            if show_tracebacks:
                 pm.displayError( traceback.format_exc())
            raise
    return result
 




def _set_docstring(attr, widget):
    script = attr.get()
    docstring = get_module_docstring(script)
    pm.control(widget, edit=True, ann=docstring)


def _active_cb_changed(attr, path_ctl):
    pm.textField(path_ctl, edit=True, enable=attr.get())


def _remove_scraper(attribute):
    pm.removeMultiInstance(attribute, b=True)



import glob
import imp
import importlib
import json
import os
import re
import pymel.core as pm

from ciocore import data as coredata
from ciocore.expander import Expander
from ciocore.gpath import Path
from ciocore.gpath_list import GLOBBABLE_REGEX, PathList
from ciocore.package_environment import PackageEnvironment
from ciocore.sequence import Sequence

import maya.api.OpenMaya as om

from ciomaya.lib import const as k
from ciomaya.lib import layer_utils


def maya_useNewAPI():
    pass


class conductorRender(om.MPxNode):
    # pass

    # static attributes
    aTitle = None

    aChunkSize = None
    aUseCustomRange = None
    aCustomRange = None
    aStartFrame = None
    aEndFrame = None
    aByFrame = None
    aAnimation = None
    aUseScoutFrames = None
    aScoutFrames = None

    aTaskTemplate = None

    aInstanceTypeName = None
    aPreemptible = None
    aProjectName = None
    aRenderLayers = None
    aCurrentRenderLayer = None

    aHostSoftware = None
    aPluginSoftware = None
    aExtraAssets = None

    aAssetScraperName = None
    aAssetScraperActive = None
    aAssetScrapers = None

    aExtraEnvironment = None
    aExtraEnvironmentKey = None
    aExtraEnvironmentValue = None
    aExtraEnvironmentExclusive = None

    aMetadata = None
    aMetadataKey = None
    aMetadataValue = None

    aUploadOnly = None
    aUseUploadDaemon = None

    aEmailAddresses = None
    aEmailAddress = None
    aEmailAddressActive = None

    aRetriesWhenPreempted = None
    aRetriesWhenFailed = None

    aTaskLimit = None
    aDoScrape = None

    aFrameCount = None
    aTaskCount = None
    aScoutTaskCount = None
    aAssetCount = None
    aAssetsSize = None
    aFrameSpec = None
    aScoutSpec = None
    aCurrentRenderer = None
    aCurrentTime = None

    aDestinationDirectory = None
    aLocationTag = None

    aAutosave = None
    aAutosaveTemplate = None
    aCleanupAutosave = None

    aShowTracebacks = None
    aFixturesDirectory = None
    aUseFixtures = None

    aOutput = None

    id = om.MTypeId(0x880500)

    @staticmethod
    def creator():
        return conductorRender()

    @classmethod
    def initialize(cls):
        cls.make_title_att()
        cls.make_frames_atts()
        cls.make_instance_type_att()
        cls.make_project_name_att()
        cls.make_layer_atts()
        cls.make_software_att()
        cls.make_assets_atts()
        cls.make_environment_atts()
        cls.make_task_atts()
        cls.make_upload_flag_atts()
        cls.make_notification_atts()
        cls.make_metadata_atts()
        cls.make_retries_atts()

        cls.make_hidden_atts()
        cls.make_info_atts()

        cls.make_autosave_atts()
        cls.make_developer_atts()
        cls.make_misc_atts()

        cls.make_output_att()

        cls.setup_attribute_affects()

    @staticmethod
    def _make_output_int_att(longname, shortname):
        nAttr = om.MFnNumericAttribute()
        att = nAttr.create(longname, shortname, om.MFnNumericData.kInt)
        nAttr.storable = False
        nAttr.writable = False
        nAttr.readable = True
        om.MPxNode.addAttribute(att)
        return att

    @classmethod
    def make_info_atts(cls):

        cls.aFrameCount = cls._make_output_int_att("frameCount", "frc")
        cls.aTaskCount = cls._make_output_int_att("taskCount", "tsc")
        cls.aScoutTaskCount = cls._make_output_int_att("scoutTaskCount", "stc")
        cls.aAssetCount = cls._make_output_int_att("assetCount", "asc")
        cls.aAssetsSize = cls._make_output_int_att("assetsSize", "asz")

        tAttr = om.MFnTypedAttribute()
        cls.aFrameSpec = tAttr.create("frameSpec", "fms", om.MFnData.kString)
        tAttr.storable = False
        tAttr.writable = False
        tAttr.readable = True
        om.MPxNode.addAttribute(cls.aFrameSpec)

        cls.aScoutSpec = tAttr.create("scoutSpec", "scs", om.MFnData.kString)
        tAttr.storable = False
        tAttr.writable = False
        tAttr.readable = True
        om.MPxNode.addAttribute(cls.aScoutSpec)

    @classmethod
    def make_title_att(cls):
        tAttr = om.MFnTypedAttribute()
        cls.aTitle = tAttr.create("title", "ttl", om.MFnData.kString)
        tAttr.hidden = False
        tAttr.storable = True
        tAttr.readable = True
        tAttr.writable = True
        om.MPxNode.addAttribute(cls.aTitle)

    @classmethod
    def make_instance_type_att(cls):
        tAttr = om.MFnTypedAttribute()
        nAttr = om.MFnNumericAttribute()
        cls.aInstanceTypeName = tAttr.create(
            "instanceTypeName", "itn", om.MFnData.kString
        )
        tAttr.hidden = False
        tAttr.storable = True
        tAttr.readable = True
        tAttr.writable = True
        om.MPxNode.addAttribute(cls.aInstanceTypeName)

        cls.aPreemptible = nAttr.create(
            "preemptible", "prm", om.MFnNumericData.kBoolean, True
        )
        nAttr.keyable = False
        nAttr.storable = True
        nAttr.readable = True
        nAttr.writable = True
        om.MPxNode.addAttribute(cls.aPreemptible)

    @classmethod
    def make_project_name_att(cls):
        tAttr = om.MFnTypedAttribute()
        cls.aProjectName = tAttr.create(
            "projectName", "prn", om.MFnData.kString
        )
        tAttr.hidden = False
        tAttr.storable = True
        tAttr.readable = True
        tAttr.writable = True
        om.MPxNode.addAttribute(cls.aProjectName)

    @classmethod
    def make_layer_atts(cls):
        eAttr = om.MFnEnumAttribute()
        nAttr = om.MFnNumericAttribute()
        tAttr = om.MFnTypedAttribute()

        cls.aRenderLayers = eAttr.create(
            "renderLayers", "rl", k.CURRENT_LAYER
        )
        eAttr.addField("current", k.CURRENT_LAYER)
        eAttr.addField("renderable", k.RENDERABLE_LAYERS)
        eAttr.hidden = False
        eAttr.keyable = True
        eAttr.storable = True
        om.MPxNode.addAttribute(cls.aRenderLayers)

        cls.aCurrentRenderLayer = nAttr.create(
            "currentRenderLayer", "crl", om.MFnNumericData.kInt
        )
        nAttr.hidden = False
        nAttr.writable = True
        nAttr.keyable = True
        nAttr.storable = False
        om.MPxNode.addAttribute(cls.aCurrentRenderLayer)

        cls.aCurrentRenderer = tAttr.create(
            "currentRenderer", "cren", om.MFnData.kString
        )
        tAttr.hidden = True
        tAttr.storable = True
        tAttr.readable = True
        tAttr.writable = True
        om.MPxNode.addAttribute(cls.aCurrentRenderer)

    @classmethod
    def make_software_att(cls):
        tAttr = om.MFnTypedAttribute()
        cls.aHostSoftware = tAttr.create(
            "hostSoftware", "hsw", om.MFnData.kString
        )
        tAttr.hidden = False
        tAttr.writable = True
        om.MPxNode.addAttribute(cls.aHostSoftware)

        cls.aPluginSoftware = tAttr.create(
            "pluginSoftware", "psw", om.MFnData.kString
        )
        tAttr.array = True
        tAttr.hidden = False
        tAttr.writable = True
        om.MPxNode.addAttribute(cls.aPluginSoftware)

    @classmethod
    def make_assets_atts(cls):
        cAttr = om.MFnCompoundAttribute()
        tAttr = om.MFnTypedAttribute()
        nAttr = om.MFnNumericAttribute()

        cls.aAssetScraperName = tAttr.create(
            "assetScraperName", "asn", om.MFnData.kString
        )

        cls.aAssetScraperActive = nAttr.create(
            "assetScraperActive", "asa", om.MFnNumericData.kBoolean, True
        )

        cls.aAssetScrapers = cAttr.create("assetScrapers", "ascs")
        cAttr.array = True
        cAttr.hidden = False
        cAttr.writable = True
        cAttr.addChild(cls.aAssetScraperName)
        cAttr.addChild(cls.aAssetScraperActive)
        om.MPxNode.addAttribute(cls.aAssetScrapers)

        conductorRender.aExtraAssets = tAttr.create(
            "extraAssets", "eass", om.MFnData.kString
        )
        tAttr.array = True
        tAttr.hidden = False
        tAttr.writable = True
        tAttr.usedAsFilename = True
        om.MPxNode.addAttribute(conductorRender.aExtraAssets)

    @classmethod
    def make_environment_atts(cls):
        cAttr = om.MFnCompoundAttribute()
        tAttr = om.MFnTypedAttribute()
        nAttr = om.MFnNumericAttribute()
        cls.aExtraEnvironmentKey = tAttr.create(
            "extraEnvironmentKey", "eek", om.MFnData.kString
        )
        cls.aExtraEnvironmentValue = tAttr.create(
            "extraEnvironmentValue", "eev", om.MFnData.kString
        )
        cls.aExtraEnvironmentExclusive = nAttr.create(
            "extraEnvironmentExclusive", "eee", om.MFnNumericData.kBoolean, False
        )
        cls.aExtraEnvironment = cAttr.create("extraEnvironment", "een")

        cAttr.hidden = False
        cAttr.writable = True
        cAttr.array = True
        cAttr.addChild(cls.aExtraEnvironmentKey)
        cAttr.addChild(cls.aExtraEnvironmentValue)
        cAttr.addChild(cls.aExtraEnvironmentExclusive)
        om.MPxNode.addAttribute(cls.aExtraEnvironment)

    @classmethod
    def make_metadata_atts(cls):
        cAttr = om.MFnCompoundAttribute()
        tAttr = om.MFnTypedAttribute()
        cls.aMetadataKey = tAttr.create(
            "metadataKey", "mdk", om.MFnData.kString
        )
        cls.aMetadataValue = tAttr.create(
            "metadataValue", "mdv", om.MFnData.kString
        )

        cls.aMetadata = cAttr.create("metadata", "md")

        cAttr.hidden = False
        cAttr.writable = True
        cAttr.array = True
        cAttr.addChild(cls.aMetadataKey)
        cAttr.addChild(cls.aMetadataValue)
        om.MPxNode.addAttribute(cls.aMetadata)

    @classmethod
    def make_retries_atts(cls):
        nAttr = om.MFnNumericAttribute()

        cls.aRetriesWhenFailed = nAttr.create(
            "retriesWhenFailed", "rwf", om.MFnNumericData.kInt, 1
        )
        nAttr.writable = True
        nAttr.keyable = True
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aRetriesWhenFailed)

        cls.aRetriesWhenPreempted = nAttr.create(
            "retriesWhenPreempted", "rwp", om.MFnNumericData.kInt, 1
        )
        nAttr.writable = True
        nAttr.keyable = True
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aRetriesWhenPreempted)

    @classmethod
    def make_upload_flag_atts(cls):
        nAttr = om.MFnNumericAttribute()

        cls.aUseUploadDaemon = nAttr.create(
            "useUploadDaemon", "uud", om.MFnNumericData.kBoolean, False
        )
        nAttr.writable = True
        nAttr.keyable = True
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aUseUploadDaemon)

        conductorRender.aUploadOnly = nAttr.create(
            "uploadOnly", "upo", om.MFnNumericData.kBoolean, False
        )
        nAttr.writable = True
        nAttr.keyable = True
        nAttr.storable = True
        om.MPxNode.addAttribute(conductorRender.aUploadOnly)

    @classmethod
    def make_frames_atts(cls):
        tAttr = om.MFnTypedAttribute()
        nAttr = om.MFnNumericAttribute()
        uAttr = om.MFnUnitAttribute()

        cls.aStartFrame = uAttr.create(
            "startFrame", "stf", om.MFnUnitAttribute.kTime, 1
        )
        uAttr.writable = True
        uAttr.keyable = False
        uAttr.storable = True
        om.MPxNode.addAttribute(cls.aStartFrame)

        cls.aEndFrame = uAttr.create(
            "endFrame", "enf", om.MFnUnitAttribute.kTime, 10
        )
        uAttr.writable = True
        uAttr.keyable = False
        uAttr.storable = True
        om.MPxNode.addAttribute(cls.aEndFrame)

        cls.aByFrame = nAttr.create(
            "byFrame", "byf", om.MFnNumericData.kInt, 1
        )
        nAttr.writable = True
        nAttr.keyable = False
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aByFrame)

        cls.aAnimation = nAttr.create(
            "animation", "ani", om.MFnNumericData.kBoolean
        )
        nAttr.writable = True
        nAttr.keyable = False
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aAnimation)

        cls.aChunkSize = nAttr.create(
            "chunkSize", "csz", om.MFnNumericData.kInt, 1
        )
        nAttr.writable = True
        nAttr.keyable = False
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aChunkSize)

        cls.aUseCustomRange = nAttr.create(
            "useCustomRange", "ucr", om.MFnNumericData.kBoolean, False
        )
        nAttr.writable = True
        nAttr.keyable = False
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aUseCustomRange)

        cls.aCustomRange = tAttr.create(
            "customRange", "crn", om.MFnData.kString
        )
        tAttr.writable = True
        tAttr.storable = True
        om.MPxNode.addAttribute(cls.aCustomRange)

        cls.aUseScoutFrames = nAttr.create(
            "useScoutFrames", "usf", om.MFnNumericData.kBoolean, True
        )
        nAttr.writable = True
        nAttr.keyable = False
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aUseScoutFrames)

        cls.aScoutFrames = tAttr.create(
            "scoutFrames", "scf", om.MFnData.kString
        )
        tAttr.writable = True
        tAttr.storable = True
        om.MPxNode.addAttribute(cls.aScoutFrames)

        cls.aCurrentFrame = uAttr.create(
            "currentFrame", "cf", om.MFnUnitAttribute.kTime)
        uAttr.writable = True
        uAttr.keyable = False
        uAttr.storable = True
        om.MPxNode.addAttribute(cls.aCurrentFrame)

    @classmethod
    def make_task_atts(cls):
        tAttr = om.MFnTypedAttribute()
        cls.aTaskTemplate = tAttr.create(
            "taskTemplate", "ttm", om.MFnData.kString
        )
        tAttr.writable = True
        tAttr.storable = True
        om.MPxNode.addAttribute(cls.aTaskTemplate)

    @classmethod
    def make_notification_atts(cls):
        cAttr = om.MFnCompoundAttribute()
        tAttr = om.MFnTypedAttribute()
        nAttr = om.MFnNumericAttribute()

        cls.aEmailAddress = tAttr.create(
            "emailAddress", "eml", om.MFnData.kString
        )

        cls.aEmailAddressActive = nAttr.create(
            "emailAddressActive", "emla", om.MFnNumericData.kBoolean, True
        )

        cls.aEmailAddresses = cAttr.create("emailAddresses", "emls")

        cAttr.hidden = False
        cAttr.writable = True
        cAttr.array = True
        cAttr.addChild(cls.aEmailAddress)
        cAttr.addChild(cls.aEmailAddressActive)
        om.MPxNode.addAttribute(cls.aEmailAddresses)

    @classmethod
    def make_misc_atts(cls):
        tAttr = om.MFnTypedAttribute()
        cls.aDestinationDirectory = tAttr.create(
            "destinationDirectory", "ddr", om.MFnData.kString
        )
        tAttr.writable = True
        tAttr.storable = True
        tAttr.usedAsFilename = True
        om.MPxNode.addAttribute(cls.aDestinationDirectory)

        cls.aLocationTag = tAttr.create(
            "locationTag", "lct", om.MFnData.kString
        )
        tAttr.writable = True
        tAttr.storable = True
        om.MPxNode.addAttribute(cls.aLocationTag)

    @classmethod
    def make_developer_atts(cls):
        nAttr = om.MFnNumericAttribute()
        tAttr = om.MFnTypedAttribute()

        cls.aShowTracebacks = nAttr.create(
            "showTracebacks", "trc", om.MFnNumericData.kBoolean, False
        )
        nAttr.writable = True
        nAttr.storable = True
        nAttr.hidden = True
        om.MPxNode.addAttribute(cls.aShowTracebacks)

        cls.aUseFixtures = nAttr.create(
            "useFixtures", "ufx", om.MFnNumericData.kBoolean, False
        )
        nAttr.writable = True
        nAttr.storable = True
        # nAttr.hidden = True
        om.MPxNode.addAttribute(cls.aUseFixtures)


        cls.aFixturesDirectory = tAttr.create(
            "fixturesDirectory", "fdr", om.MFnData.kString
        )
        tAttr.writable = True
        tAttr.storable = True
        tAttr.usedAsFilename = True
        om.MPxNode.addAttribute(cls.aFixturesDirectory)



    @classmethod
    def make_autosave_atts(cls):

        tAttr = om.MFnTypedAttribute()
        nAttr = om.MFnNumericAttribute()

        cls.aAutosaveTemplate = tAttr.create(
            "autosaveTemplate", "ast", om.MFnData.kString
        )
        tAttr.writable = True
        tAttr.storable = True
        om.MPxNode.addAttribute(cls.aAutosaveTemplate)

        cls.aAutosave = nAttr.create(
            "autosave", "aus", om.MFnNumericData.kBoolean, True
        )
        nAttr.writable = True
        nAttr.storable = True
        nAttr.hidden = True
        om.MPxNode.addAttribute(cls.aAutosave)

        cls.aCleanupAutosave = nAttr.create(
            "cleanupAutosave", "cua", om.MFnNumericData.kBoolean, True
        )
        nAttr.writable = True
        nAttr.keyable = True
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aCleanupAutosave)

    @classmethod
    def make_hidden_atts(cls):
        nAttr = om.MFnNumericAttribute()
        cls.aDoScrape = nAttr.create(
            "doScrape", "dsc", om.MFnNumericData.kBoolean, False
        )
        nAttr.writable = True
        nAttr.storable = True
        nAttr.hidden = True
        om.MPxNode.addAttribute(cls.aDoScrape)

        cls.aTaskLimit = nAttr.create(
            "taskLimit", "tsl", om.MFnNumericData.kInt, 10
        )
        nAttr.writable = True
        nAttr.storable = True
        nAttr.hidden = True
        om.MPxNode.addAttribute(cls.aTaskLimit)

    @classmethod
    def make_output_att(cls):
        """
        Output atttribute.
        """
        tAttr = om.MFnTypedAttribute()
        cls.aOutput = tAttr.create("output", "out", om.MFnData.kString)
        tAttr.readable = True
        tAttr.storable = False
        tAttr.writable = False
        tAttr.keyable = False

        om.MPxNode.addAttribute(cls.aOutput)

    @classmethod
    def setup_attribute_affects(cls):
        om.MPxNode.attributeAffects(cls.aTitle, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aInstanceTypeName, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aProjectName, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aChunkSize, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aUseCustomRange, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aCustomRange, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aStartFrame, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aEndFrame, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aByFrame, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aAnimation, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aUseScoutFrames, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aScoutFrames, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aHostSoftware, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aPluginSoftware, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aExtraEnvironment, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aTaskTemplate, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aTaskLimit, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aDoScrape, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aExtraAssets, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aAssetScrapers, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aMetadata, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aRetriesWhenFailed, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aRetriesWhenPreempted, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aUseUploadDaemon, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aUploadOnly, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aEmailAddresses, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aDestinationDirectory, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aLocationTag, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aCurrentRenderLayer, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aCurrentRenderer, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aCurrentFrame, cls.aOutput)

        om.MPxNode.attributeAffects(cls.aUseCustomRange, cls.aFrameCount)
        om.MPxNode.attributeAffects(cls.aCustomRange, cls.aFrameCount)
        om.MPxNode.attributeAffects(cls.aStartFrame, cls.aFrameCount)
        om.MPxNode.attributeAffects(cls.aEndFrame, cls.aFrameCount)
        om.MPxNode.attributeAffects(cls.aByFrame, cls.aFrameCount)
        om.MPxNode.attributeAffects(cls.aAnimation, cls.aFrameCount)

        om.MPxNode.attributeAffects(cls.aChunkSize, cls.aTaskCount)
        om.MPxNode.attributeAffects(cls.aUseCustomRange, cls.aTaskCount)
        om.MPxNode.attributeAffects(cls.aCustomRange, cls.aTaskCount)
        om.MPxNode.attributeAffects(cls.aStartFrame, cls.aTaskCount)
        om.MPxNode.attributeAffects(cls.aEndFrame, cls.aTaskCount)
        om.MPxNode.attributeAffects(cls.aByFrame, cls.aTaskCount)
        om.MPxNode.attributeAffects(cls.aAnimation, cls.aTaskCount)

        om.MPxNode.attributeAffects(cls.aUseScoutFrames, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aScoutFrames, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aChunkSize, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aUseCustomRange, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aCustomRange, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aStartFrame, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aEndFrame, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aByFrame, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aAnimation, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aCurrentFrame, cls.aScoutTaskCount)
  
        om.MPxNode.attributeAffects(cls.aUseCustomRange, cls.aFrameSpec)
        om.MPxNode.attributeAffects(cls.aCustomRange, cls.aFrameSpec)
        om.MPxNode.attributeAffects(cls.aStartFrame, cls.aFrameSpec)
        om.MPxNode.attributeAffects(cls.aEndFrame, cls.aFrameSpec)
        om.MPxNode.attributeAffects(cls.aByFrame, cls.aFrameSpec)
        om.MPxNode.attributeAffects(cls.aAnimation, cls.aFrameSpec)
        om.MPxNode.attributeAffects(cls.aCurrentFrame, cls.aFrameSpec)
   
        om.MPxNode.attributeAffects(cls.aUseScoutFrames, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aScoutFrames, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aChunkSize, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aUseCustomRange, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aCustomRange, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aStartFrame, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aEndFrame, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aByFrame, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aAnimation, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aCurrentFrame, cls.aScoutSpec)


    def compute(self, plug, data):
        """Compute output json from input attribs."""
        if (not ((plug == self.aOutput) or
                 (plug == self.aFrameCount) or
                 (plug == self.aTaskCount) or
                 (plug == self.aScoutTaskCount) or
                 (plug == self.aFrameSpec)or
                 (plug == self.aScoutSpec)
                 )):
            return None

        try:
            sequence = self.get_sequence(data)
        except (ValueError, TypeError):
            om.MGlobal.displayWarning("Invalid frame sequence specified.")
            return None

        scout_sequence = self.get_scout_sequence(data, sequence)
        frame_count = len(sequence)
        task_count = sequence.chunk_count()

        scout_task_count = 0
        scout_tasks_sequence = None
        if scout_sequence:
            scout_chunks = sequence.intersecting_chunks(scout_sequence)
            if scout_chunks:
                scout_tasks_sequence = Sequence.create(
                    ",".join(str(chunk) for chunk in scout_chunks))
                scout_task_count = len(scout_chunks)

        self.set_frame_info_plugs(
            data, frame_count, task_count, scout_task_count)

        self.set_frame_and_scout_spec(data, sequence, scout_tasks_sequence)

        if (plug != self.aOutput):
            return self

        context = self.get_context(data)
        expander = Expander(**context)

        handle = data.outputValue(self.aOutput)
        result = {}
        result.update(self.get_software_environment(data))
        result.update(self.get_instance_type(data))
        result.update(self.get_title(data, expander))
        result.update(self.get_project(data))
        result.update(self.get_tasks(data, sequence, context))
        result.update(self.get_upload_paths(data))
        result.update(self.get_upload_flags(data))
        result.update(self.get_retry_policy(data))
        result.update(self.get_notifiations(data))
        result.update(self.get_scout_frames(scout_sequence))
        result.update(self.get_metadata(data, expander))
        result.update(self.get_location_tag(data))
        result.update(self.get_destination_directory(data))

        handle.setString(json.dumps(result))

        data.setClean(plug)
        return self

    @classmethod
    def get_sequence(cls, data):
        chunk_size = data.inputValue(cls.aChunkSize).asInt()
        use_custom_range = data.inputValue(cls.aUseCustomRange).asBool()
        if use_custom_range:
            custom_range = data.inputValue(cls.aCustomRange).asString()
            return Sequence.create(custom_range, chunk_size=chunk_size, chunk_strategy="progressions")

        if not data.inputValue(cls.aAnimation).asBool():

            return Sequence.create(int(data.inputValue(cls.aCurrentFrame).asTime().asUnits(om.MTime.uiUnit())))

        start_frame = data.inputValue(
            cls.aStartFrame).asTime().asUnits(om.MTime.uiUnit())
        end_frame = data.inputValue(
            cls.aEndFrame).asTime().asUnits(om.MTime.uiUnit())
        by_frame = data.inputValue(cls.aByFrame).asInt()
        return Sequence.create(int(start_frame), int(end_frame), by_frame, chunk_size=chunk_size, chunk_strategy="progressions")

    @classmethod
    def get_scout_sequence(cls, data, main_sequence):
        use_scout_frames = data.inputValue(cls.aUseScoutFrames).asBool()
        if not use_scout_frames:
            return

        scout_frames = data.inputValue(cls.aScoutFrames).asString().strip()

        match = re.compile(r"^auto[, :]+(\d+)$").match(scout_frames)
        if match:
            samples = int(match.group(1))
            return main_sequence.subsample(samples)

        try:
            return Sequence.create(scout_frames)
        except (ValueError, TypeError):
            return

    @classmethod
    def get_scout_frames(cls, scout_sequence):
        return {"scout_frames": ",".join([str(s) for s in scout_sequence or []])}

    @classmethod
    def set_frame_info_plugs(cls, data, frame_count, task_count, scout_task_count):
        handle = data.outputValue(cls.aFrameCount)
        handle.setInt(frame_count)
        handle.setClean()
        handle = data.outputValue(cls.aTaskCount)
        handle.setInt(task_count)
        handle.setClean()
        handle = data.outputValue(cls.aScoutTaskCount)
        handle.setInt(scout_task_count)
        handle.setClean()

    @classmethod
    def set_frame_and_scout_spec(cls, data, sequence, scout_sequence):
        handleFrameSpec = data.outputValue(cls.aFrameSpec)
        handleScoutSpec = data.outputValue(cls.aScoutSpec)
        if scout_sequence:
            handleScoutSpec.setString(str(scout_sequence))
        else:
            handleScoutSpec.setString("")
        handleFrameSpec.setString(str(sequence))

        handleScoutSpec.setClean()
        handleFrameSpec.setClean()

    @classmethod
    def get_instance_type(cls, data):
        return {
            "instance_type": data.inputValue(cls.aInstanceTypeName).asString(),
            "preemptible": data.inputValue(cls.aPreemptible).asBool()
        }

    @classmethod
    def get_title(cls, data, expander):
        title = data.inputValue(cls.aTitle).asString()
        return {"job_title": expander.evaluate(title)}

    @classmethod
    def get_project(cls, data):
        return {"project": data.inputValue(cls.aProjectName).asString()}

    @classmethod
    def get_software_environment(cls, data):
        extra_env = cls.get_extra_env(data)
        packages_data = cls.get_software_packages(data)
        packages_data["env"].extend(extra_env)
        return {
            "environment": dict(packages_data["env"]),
            "software_package_ids": packages_data["ids"]
        }

    @classmethod
    def get_software_packages(cls, data):
        tree_data = coredata.data().get("software")
        paths = []
        host_path = data.inputValue(cls.aHostSoftware).asString()
        paths.append(host_path)
        array_handle = data.inputArrayValue(cls.aPluginSoftware)

        while not array_handle.isDone():
            plugin_path = "{}/{}".format(host_path,
                                         array_handle.inputValue().asString())
            paths.append(plugin_path)
            array_handle.next()

        result = {
            "ids": [],
            "env": PackageEnvironment()

        }

        for package in filter(None, [tree_data.find_by_path(path) for path in paths if path]):
            if package:
                result["ids"].append(package["package_id"])
                result["env"].extend(package)

        return result

    @classmethod
    def get_extra_env(cls, data):
        result = []
        array_handle = data.inputArrayValue(cls.aExtraEnvironment)
        while not array_handle.isDone():
            name = array_handle.inputValue().child(
                cls.aExtraEnvironmentKey).asString()
            value = array_handle.inputValue().child(
                cls.aExtraEnvironmentValue).asString()
            exclusive = array_handle.inputValue().child(
                cls.aExtraEnvironmentExclusive).asBool()
            name = name.strip()
            value = value.strip()

            if name and value:
                result.append(
                    {
                        "name": name,
                        "value": value,
                        "merge_policy": "exclusive" if exclusive else "append"
                    }
                )
            array_handle.next()
        return result

    def get_context(self, data):
        node = om.MFnDependencyNode(self.thisMObject())
        file_name = pm.sceneName()
        scene_name = os.path.splitext(os.path.split(file_name)[1])[0]

        context = {
            "Scene": "undefined",
            "SceneFile": "undefined",
            "Object": node.name(),
            "RenderLayer": layer_utils.get_current_layer_name(),
            "WorkspacePath": "undefined",
            "OutputPath":  "undefined",
            "Renderer": data.inputValue(self.aCurrentRenderer).asString(),
            "ConductorVersion": pm.moduleInfo(version=True, moduleName="conductor")
        }

        if file_name:
            context["Scene"] = scene_name
            context["SceneFile"] = Path(file_name).posix_path(with_drive=False)
        try:
            context["WorkspacePath"] = Path(pm.workspace(
                q=True, rd=True)).posix_path(with_drive=False)
        except BaseException:
            pass

        try:
            context["OutputPath"] = Path(data.inputValue(
                self.aDestinationDirectory).asString().strip()).posix_path(with_drive=False)
        except BaseException:
            pass

        low_context = {}
        for key in context:
            key_lower = key.lower()
            if not key_lower in context:
                low_context[key_lower] = context[key]
        context.update(low_context)
        return context

    @classmethod
    def get_tasks(cls, data, sequence, context):
        if data.inputValue(cls.aUploadOnly).asBool():
            return {}

        tasks = []
        template = data.inputValue(cls.aTaskTemplate).asString()
        limit = data.inputValue(cls.aTaskLimit).asInt()
        chunks = sequence.chunks()
        if limit < 0:
            limit = len(chunks)
        for i, chunk in enumerate(chunks):
            if i >= limit:
                break
            task_context = {
                "start": str(chunk.start),
                "end": str(chunk.end),
                "step": str(chunk.step),
                "chunk_length":  str(len(chunk))
            }
            task_context.update(context)
            expander = Expander(**task_context)

            tasks.append({
                "command": expander.evaluate(template),
                "frames": str(chunk)
            })
        return {"tasks_data": tasks}

    @classmethod
    def get_upload_flags(cls, data):
        return {
            "upload_only": data.inputValue(cls.aUploadOnly).asBool(),
            "local_upload": not data.inputValue(cls.aUseUploadDaemon).asBool()
        }

    @classmethod
    def get_retry_policy(cls, data):
        return {
            "autoretry_policy": {
                "preempted": {"max_retries":  data.inputValue(cls.aRetriesWhenPreempted).asInt()},
                "failed":    {"max_retries":   data.inputValue(cls.aRetriesWhenFailed).asInt()}
            }
        }

    @classmethod
    def get_notifiations(cls, data):
        result = []
        array_handle = data.inputArrayValue(cls.aEmailAddresses)
        while not array_handle.isDone():
            if array_handle.inputValue().child(
                    cls.aEmailAddressActive).asBool():
                value = array_handle.inputValue().child(
                    cls.aEmailAddress).asString().strip()
            if value:
                result.append(value)
            array_handle.next()
        return {"notify": result}

    def get_upload_paths(self, data):
        node_name = om.MFnDependencyNode(self.thisMObject()).name()

        path_list = PathList()
        path_list.add(*self.get_scraped_paths(data, node_name))
        path_list.add(*self.get_cached_paths(data))
        path_list.remove_missing()
        path_list.glob()
        return {"upload_paths": sorted([p.posix_path() for p in path_list])}

    @classmethod
    def get_scraped_paths(cls, data, node):
        result = []
        if not data.inputValue(cls.aDoScrape).asBool():
            return result
        array_handle = data.inputArrayValue(cls.aAssetScrapers)
        while not array_handle.isDone():
            if array_handle.inputValue().child(cls.aAssetScraperActive).asBool():
                script = array_handle.inputValue().child(cls.aAssetScraperName).asString()
                try:
                    scraper_module = importlib.import_module(script)
                    reload(scraper_module)
                    result += scraper_module.run(node)
                except SyntaxError:
                    om.MGlobal.displayWarning(
                        "Syntax error in the scraper script: '{}'.".format(script))
                    raise
                except ImportError:
                    om.MGlobal.displayWarning(
                        "Can't load the script '{}' as a Python module. Skipping!!".format(script))
                except BaseException:
                    om.MGlobal.displayWarning(
                        "Unknown problem with scraper: '{}'. Skipping!!".format(script))
                
            array_handle.next()

        return [p["path"] for p in result]

    @classmethod
    def get_cached_paths(cls, data):
        result = []
        array_handle = data.inputArrayValue(cls.aExtraAssets)
        while not array_handle.isDone():
            path = array_handle.inputValue().asString().strip()
            if path:
                result.append(path)
            array_handle.next()
        return result

    @classmethod
    def get_metadata(cls, data, expander):
        metadata = {}
        array_handle = data.inputArrayValue(cls.aMetadata)
        while not array_handle.isDone():
            key = array_handle.inputValue().child(
                cls.aMetadataKey).asString().strip()
            value = array_handle.inputValue().child(
                cls.aMetadataValue).asString().strip()

            metadata[key] = value

            array_handle.next()

        return {"metadata": expander.evaluate(metadata)}

    @classmethod
    def get_location_tag(cls, data):
        result = data.inputValue(cls.aLocationTag).asString().strip()
        return {"location": result} if result else {}

    @classmethod
    def get_destination_directory(cls, data):
        return {"output_path": data.inputValue(cls.aDestinationDirectory).asString().strip()}



# from __future__ import unicode_literals

# import argparse
import platform
import os
import re
import sys
import errno
from shutil import copy2

# /users/me/Conductor/ciomaya
PKG_DIR = os.path.dirname(os.path.abspath(__file__))
MODULE_DIR = PKG_DIR
PKGNAME = os.path.basename(PKG_DIR)  # ciomaya
MODULE_FILENAME = "conductor.mod"

PLATFORM = sys.platform

def main():
    if not PLATFORM in ["darwin", "win32", "linux"]:
        sys.stderr.write("Unsupported platform: {}".format(PLATFORM))
        sys.exit(1)

    module_dirs = get_maya_module_dirs()
    transform_maya_mod_file(module_dirs)
    sys.stdout.write("Completed Maya Module setup!\n")


def get_maya_module_dirs():
    """Get places to write a modules file."""

    if PLATFORM == "darwin":
        app_dir = "~/Library/Preferences/Autodesk/maya"
    elif PLATFORM == "linux":
        app_dir = "~/maya"
    else:  # windows
        app_dir = "~\Documents\maya"

    # If Maya app dir is set, add it
    mod_dirs = filter(None, [os.environ.get(
        "MAYA_APP_DIR"), os.path.expanduser(app_dir)])

    return [os.path.join(d, "modules") for d in mod_dirs]


def transform_maya_mod_file(module_paths):

    mod_file = os.path.join(MODULE_DIR, MODULE_FILENAME)

    lines = []
    with open(mod_file) as f:
        for i, line in enumerate(f):
            if i == 0:
                lines.append(re.sub(r" \.$",  " {}".format(MODULE_DIR), line))
            else:
                lines.append(line)

    for dest in module_paths:
        try:
            ensure_directory(dest)
        except BaseException:
            sys.stderr.write("Invalid directory: {}\n".format(dest))
            continue

        fn = os.path.join(dest, MODULE_FILENAME)
        with open(fn, "w") as f:
            for line in lines:
                f.write(unicode(line))
        sys.stdout.write("Wrote Maya module file: {}\n".format(fn))


def ensure_directory(directory):
    try:
        os.makedirs(directory)
    except OSError as ex:
        if ex.errno == errno.EEXIST and os.path.isdir(directory):
            pass
        else:
            raise

if __name__ == '__main__':
    main()



"""
A scraper to collect paths from Yeti nodes.
"""

import pymel.core as pm
import re

from ciocore.gpath_list import GLOBBABLE_REGEX,PathList
from ciocore.gpath import Path
from ciomaya.lib import scraper_utils
from ciomaya.lib import software
import os


YETI_INTERNAL_NODE_TYPES = {
    "texture": ["file_name"],
    "reference": ["reference_file"]
}

PADDED_PRINTF_REGEX = r"%0\d+d"

def run(_):

    if not software.detect_yeti():
        return []

    paths = sum([_scrape_yeti_graph_node(yeti_graph_node)
                 for yeti_graph_node in pm.ls(type="pgYetiMaya")], [])

    paths = scraper_utils.starize_tokens(paths, PADDED_PRINTF_REGEX)
    return paths


def _scrape_yeti_graph_node(graph_node):
    """
    Get the files
    """
    paths = []
    is_caching = graph_node.attr("fileMode").get()
    if is_caching:
        plug = graph_node.attr("cacheFileName")
        cache_path = plug.get()
        if cache_path:
            paths.append({"path": cache_path, "plug": plug})

    search_paths = _get_search_paths(graph_node)
    for yeti_node_type in YETI_INTERNAL_NODE_TYPES:
        attrs = YETI_INTERNAL_NODE_TYPES[yeti_node_type]
        yeti_nodes = pm.pgYetiGraph(
            graph_node, listNodes=True, type=yeti_node_type)
        for yeti_node in yeti_nodes:
            for attr in attrs:

                possible_paths = _possible_asset_paths(graph_node,
                                               yeti_node, attr, search_paths)

                paths+= possible_paths

    return paths


def _possible_asset_paths(graph_node, yeti_node, attr, search_paths):

    asset_path = pm.pgYetiGraph(
        graph_node, node=yeti_node, getParamValue=True, param=attr)

    paths = []
    yeti_metadata = {
        "yeti_plug": "{}.{}".format(yeti_node, attr),
        "yeti_graph_node": graph_node
    }

    if asset_path:
        # THIS NEEDS AN EXPLANATION! 
        # Firstly, the absolute path may exist, but its also possible that it
        # doesn't and that instead the file exists in one of the search paths.
        # Therefore, we may as well treat it like a relative path and simply add
        # the directory portion to the list of search paths.
        asset_path = Path(asset_path)
        if asset_path.absolute:
            search_paths.append(os.path.dirname(asset_path.posix_path()))
        relpath = os.path.basename(asset_path.posix_path())

        # Now we are searching for a relative path in a list of search paths.

        # We prefer to avoid hitting the filesystem. Therefore we simply
        # generate a path for every combination of searchpath/relpath.

        # What happens when some of them (as expected) are eventually not found?
        # Well the policy for all scrapers is that the result of the scrape may
        # contain some explicit paths and some paths with glob characters. Glob
        # characters cause a path to be silently ignored if it is not found when
        # globbed. 
        
        # Explicit paths that are not found on the other hand will cause a
        # warning on submission.
        
        # You are thinking, how will the customer know if the file is really
        # missing by accident? Well, YETI complains if it can't find a file, so
        # the customer will already know.

        # GLOBBABLE is defined as: it contains some characters that
        # indicate it can be globbed. e.g. "[", "*", "?"
        
        # We can exploit this and make sure that all paths are
        # globbable. To do that, we wrap the last letter of each path in square
        # brackets. It will be globbed, but there can only be one path that
        # matches, similar to the way we find tx files.

        # In other words /path/to/file.01.ex[r] is the same as saying
        # /path/to/file.01.exr if it's there (fail silently).

        # If the path already contains a YETI frame number pattern, then we
        # don't need to make the path globbable here, since the pattern will be
        # replaced with a "*" anyway.


        for search_path in search_paths:
            possible_path = os.path.join(
                search_path, relpath)
            if not re.search(PADDED_PRINTF_REGEX, possible_path):
                possible_path = "{}[{}]".format(
                    possible_path[:-1], possible_path[-1])
                pathobj = {"path": possible_path}
                pathobj.update(yeti_metadata)
            paths.append(pathobj)
    return paths


def _get_search_paths(yeti_graph_node):
    path_list = PathList()
    search_paths = os.pathsep.join(
        filter(None, 
        [yeti_graph_node.attr("imageSearchPath").get() or None,  os.environ.get("PG_IMAGE_PATH")]
    ))
    
    if search_paths:
        path_list.add(*[p.strip() for p in search_paths.split(os.pathsep)])
    return [p.posix_path() for p in path_list]




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import glob
import io

import setuptools
from PIL import Image
from setuptools.command.build_py import build_py

NAME = "ciomaya"
DESCRIPTION = "Maya plugin for Conductor Cloud Rendering Platform."
URL = "https://github.com/AtomicConductor/conductor-maya"
EMAIL = "info@conductortech.com"
AUTHOR = "conductor"
REQUIRES_PYTHON = "~=2.7"
REQUIRED = ["ciocore>=0.2.15,<1.0.0"]
HERE = os.path.abspath(os.path.dirname(__file__))
DEV_VERSION = "dev.999"

with open(os.path.join(HERE, 'VERSION')) as version_file:
    VERSION = version_file.read().strip()

long_description = ""

with open(os.path.join(HERE, 'README.md')) as readme:
    long_description = readme.read().strip()

long_description += "\n\n## Changelog\n\n"
with open(os.path.join(HERE, 'CHANGELOG.md')) as changelog:
    long_description += changelog.read().strip()   
 



class BuildCommand(build_py):
    def run(self):
        build_py.run(self)
        if not self.dry_run:
            target_dir = os.path.join(self.build_lib, NAME)
            self.write_version_info(target_dir)
            self.convert_icons(target_dir)

    def write_version_info(self, target_dir):

        for filename in [
            os.path.join(target_dir, "plug-ins", "Conductor.py"),
            os.path.join(target_dir, "conductor.mod"),
        ]:
            with open(filename, "r") as f:
                filedata = f.read()
            filedata = filedata.replace(DEV_VERSION, VERSION)
            with open(filename, "w") as f:
                f.write(filedata)

        with open(os.path.join(target_dir, "VERSION"), "w") as f:
            f.write(VERSION)


    def convert_icons(self, target_dir):
        out_icon_dir = os.path.join(target_dir, "icons")
        icons_dir = os.path.join("images", "icons")
        icon_files = os.listdir(icons_dir)
        for icon_file in icon_files:
            base, ext = os.path.splitext(icon_file)
            if not ext in [".png", ".jpg"]:
                continue
            icon_path = os.path.join(icons_dir, icon_file)
            for size in [36, 30, 24, 18]:
                icon = Image.open(icon_path)
                icon.thumbnail((size, size), Image.BICUBIC)
                out_icon_file = os.path.join(
                    out_icon_dir, "{}_{}x{}{}".format(base, size, size, ext))
                icon.save(out_icon_file)

        icons_dir = os.path.join("images", "nodeicons")
        icon_files = os.listdir(icons_dir)
        for icon_file in icon_files:
            base, ext = os.path.splitext(icon_file)
            if not ext in [".png", ".jpg"]:
                continue
            icon_path = os.path.join(icons_dir, icon_file)
            for size, prefix in [(36, ""), (18, "out_")]:
                icon = Image.open(icon_path)
                icon.thumbnail((size, size), Image.NEAREST)
                out_icon_file = os.path.join(
                    out_icon_dir, "{}{}{}".format(prefix, base, ext))
                icon.save(out_icon_file)


setuptools.setup(
    author=AUTHOR,
    author_email=EMAIL,
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python",
        "Topic :: Multimedia :: Graphics :: 3D Rendering",
    ],
    cmdclass={"build_py": BuildCommand},
    description=DESCRIPTION,
    install_requires=REQUIRED,
    long_description=long_description,
    long_description_content_type="text/markdown",
    name=NAME,
    package_dir={"": "."},
    packages=setuptools.find_packages(where="."),
    include_package_data=True,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    version=VERSION,
    zip_safe=False,
    package_data={NAME: ["plug-ins/*.*", "icons/*.*",
                         "scripts/*.*", "conductor.mod"]}
)



from contextlib import contextmanager
import pymel.core as pm


def print_setAttr_cmd(attr):
    if attr.type() == "string":
        print 'setAttr -type "string" "{}" "{}";'.format(
            attr.name(), attr.get())
    else:
        print 'setAttr "{}" {};'.format(attr.name(), attr.get())


@contextmanager
def ae_template():
    pm.setUITemplate("attributeEditorTemplate", pushTemplate=True)
    yield
    pm.setUITemplate(ppt=True)



def ensure_populate_menu(menu, values):
    items = pm.optionMenu(menu, q=True, itemListLong=True)
    labels = [pm.menuItem(i, q=True, label=True) for i in items]

    if labels == values:
        return
    for item in items:
        pm.deleteUI(item)
    pm.setParent(menu, menu=True)
    for value in values:
        pm.menuItem(label=value)


def next_available_element_plug(array_plug):

    indices = array_plug.getArrayIndices()
    next_available = next(a for a, b in enumerate(indices + [-1]) if a != b)
    return array_plug[next_available]


def sync_menu_to_attr(menu, attr, **kw):
    """
    Make sure menu item reflects the attribute value.

    If the attribute is invalid, set it to the default.

    kw[value_label_pairs] is a map of corresponence from  attribute values to
    menu labels. Example:

    [
        ("n2-standard-16", "2 core, 16 Gb Ram"),
        ("n8-standard-64", "8 core, 64 Gb Ram")
    ]

    default is the index of the  value/menu-item to use in case the attribute
    value is not in the available list.
    """

    default = kw.get("default", 0)
    items = pm.optionMenu(menu, q=True, itemListLong=True)
    if not items:
        return
    value = attr.get()

    # negative index means count value back from the end.
    # TODO find a more compact way to validate
    if default < 0:
        default = len(items) + default
    if default < 0:
        default = 0

    values = kw.get("values", [pm.menuItem(item, q=True, label=True) for item in items])

    if len(values) != len(items):
        raise ValueError(
            "Values were provided but do not match the number of menu items {}/{}.".format( len(values),len(items)))
 
    try:
        index = values.index(value)
    except ValueError:
        if kw.get("immutable"):
            raise ValueError("Value {} not found in menu".format(value))
        index = default
        attr.set(values[index])
        pm.displayInfo("{} is not available. Switching to {}. ".format(
            value, values[index]))
        print_setAttr_cmd(attr)

    pm.optionMenu(menu, edit=True, sl=(index + 1))




def find_ui(name, root=None):
    """""breadth first search for element by name under root."""
    if not root:
        root = pm.setParent(q=True)
    stack = [root]
    while stack:
        if stack[0].rpartition("|")[2] == name:
            return stack[0]
        if pm.layout(stack[0], q=True, exists=True):
            for child_ui in pm.layout(stack[0], q=True, childArray=True) or[]:
                 #must ensure short name because we build full path manually
                child_ui = child_ui.rpartition("|")[2]
                stack.append("{}|{}".format(stack[0],child_ui))

        stack = stack[1:]



def find_layout(parent, layout_name):
    return [p for p in pm.layout(parent, q=True, childArray=True) if p.rpartition(
        "|")[2] == layout_name][0]



"""
A scraper to recursively collect paths from Arnold standin nodes.
"""
from ciomaya.lib import scraper_utils
import re
import arnold
import pymel.core as pm
from ciocore.sequence import Sequence
from contextlib import contextmanager


@contextmanager
def arnold_context():
    """Perform some ops between arnold begin end tags"""
    arnold.AiBegin()
    arnold.AiMsgSetConsoleFlags(arnold.AI_LOG_ALL)
    try:
        yield
    finally:
        arnold.AiEnd()


def run(node):
    """
    Recursive ass scrape.

    Can be slow, so we optimize as much as possible.
    """
    paths = []
    for path in list(set(
        [p["path"] for p in scraper_utils.get_paths({"mtoa": {"aiStandIn": ["dso"]}})]
    )):

        # Since scanning ass files recursively is expensive, we try to use only
        # those reelevant for the frame range.
        sequence = Sequence.create(pm.PyNode(node).attr("frameSpec").get())
        resolved_ass_filenames = scraper_utils.resolve_to_sequence(
            path, sequence)

        paths.extend(resolved_ass_filenames)

    paths = scraper_utils.expand_workspace(paths)

    # maintain a list of filenames to skip
    seen = set()
    found_files = []
    for path in paths:
        found_files.extend(_files_in(path, seen))

    found_files.extend(paths)
    found_files = list(set(found_files))

    return [{"path": p} for p in found_files]


def _files_in(ass_file, seen, depth=0):

    if ass_file in seen:
        return []
    seen.add(ass_file)

    print ".."*depth, ass_file
    found_ass_files = []
    found_leaf_files = []
    with arnold_context():
        arnold.AiASSLoad(ass_file, arnold.AI_NODE_ALL)
        iterator = arnold.AiUniverseGetNodeIterator(
            arnold.AI_NODE_SHAPE | arnold.AI_NODE_SHADER)
        while not arnold.AiNodeIteratorFinished(iterator):
            node = arnold.AiNodeIteratorGetNext(iterator)
            node_entry = arnold.AiNodeGetNodeEntry(node)
            node_entry_name = arnold.AiNodeEntryGetName(node_entry)

            if node_entry_name == "procedural":
                fn = arnold.AiNodeGetStr(node, "filename")
                if fn:
                    if fn.endswith(".ass"):
                        found_ass_files.append(fn)
                    else:
                        found_leaf_files.append(fn)
            elif node_entry_name == "image":
                fn = arnold.AiNodeGetStr(node, "filename")
                if fn:
                    paths = list(set(_expand_attr_token(fn)))
                    found_leaf_files.extend(paths)

        arnold.AiNodeIteratorDestroy(iterator)

    result = found_ass_files + found_leaf_files

    # recurse
    depth += 1
    for found_ass_file in found_ass_files:
        result += _files_in(found_ass_file, seen, depth)

    return result


def _expand_attr_token(filename):
    """
    Resolve filenames from mtoa user data attributes.

    Find shapes that have the attribute in the token. Get the value and expand
    the template with that name. Also expand with the default val.
    """
    result = []
    match = scraper_utils.extract_attr_token(filename)
    if not match:
        return [filename]

    template, attr_name, default_val = match

    if default_val:
        result.append(template.replace(scraper_utils.PLACEHOLDER, default_val))

    iterator = arnold.AiUniverseGetNodeIterator(arnold.AI_NODE_SHAPE)
    while not arnold.AiNodeIteratorFinished(iterator):
        node = arnold.AiNodeIteratorGetNext(iterator)
        node_entry = arnold.AiNodeGetNodeEntry(node)
        node_entry_name = arnold.AiNodeEntryGetName(node_entry)

        if node_entry_name == "polymesh":
            attr_val = arnold.AiNodeGetStr(node, attr_name)
            result.append(template.replace(
                scraper_utils.PLACEHOLDER, attr_val))

    arnold.AiNodeIteratorDestroy(iterator)

    return result



"""
A scraper to collect paths from Arnold nodes except standins.
"""

from ciomaya.lib import scraper_utils
import re
import arnold
import pymel.core as pm
from ciocore.gpath_list import GLOBBABLE_REGEX, PathList

from contextlib import contextmanager

ATTRS = {
    "mtoa":
    {
        "aiImage": ["filename"],
        "aiPhotometricLight":  ["aiFilename"],
        "aiVolume": ["dso", "filename"]
    },
    # Arnold attributes on Maya nodes
    "MayaBuiltin": {
        "file": ["computedFileTextureNamePattern"],
        "mesh": ["dso"]
    }
}

# See https://docs.arnoldrenderer.com/display/A5AFMUG/Tokens
TOKENS = (r"<tile>", r"<udim>", r"<frame>", r"<f\d?>", r"<aov>", r"#+")

 
def run(_):

    paths = scraper_utils.get_paths(ATTRS)
    paths = scraper_utils.starize_tokens(paths, *TOKENS)
    paths = _resolve_attr_tokens(paths)
    paths = scraper_utils.expand_workspace(paths)

    paths = scraper_utils.extend_with_tx_paths(paths)
    return paths
     

def _resolve_attr_tokens(paths):
    """
    Resolve image paths specified like so:

    /Volumes/xtr/gd/standin_fixture//sourceimages/<attr:floormap default:alcazar>.jpg

    # https://docs.arnoldrenderer.com/pages/viewpage.action?pageId=40110953

    # REGEX TESTER
    # https://regex101.com/r/eFp4RT/1/
    """

    result = []
    for path in paths:
        # Don't expand if no plug info
        try:
            plug = pm.Attribute(path["plug"])
        except (pm.MayaNodeError, pm.MayaAttributeError):
            result.append(path)
            continue

        # Don't expand  if not file or aiImage
        if not plug.node().type() in [u"file", u"aiImage"]:
            result.append(path)
            continue

        # Don't expand if no <attr: ... > token
        match = scraper_utils.extract_attr_token(path["path"])
        if not match:
            result.append(path)
            continue



        template, attr_name, default_val = match

        found_paths = _expand_attr_token(
            template, plug, attr_name, default_val)
        result.extend(found_paths)
    return result


def _expand_attr_token(template, plug, attr_name, default_val):
    """
    Resolve filenames from mtoa user data attributes.

    Look downstream from the texture plug to find shading engines it affects.
    Loop through the shapes shaded by those engines, testing for the attribute
    corresponding to obj_attr_name. If found, expand the template with that
    name. Also add the shape name to the result object.
    """
    mtoa_const_attr_name = "mtoa_constant_{}".format(attr_name)
    result = []
    if default_val:
        result.append({
            "path": template.replace(scraper_utils.PLACEHOLDER, default_val),
            "plug": plug.name()
        })

    for shading_engine in [node for node in pm.listHistory(plug, future=True) if node.type() == "shadingEngine"]:
        for shape in pm.sets(shading_engine, q=True):
            try:
                replacement = shape.attr(mtoa_const_attr_name).get()
            except pm.MayaAttributeError:
                continue
            result.append({
                "path": template.replace(scraper_utils.PLACEHOLDER, replacement),
                "plug": plug.name(),
                "shape": shape.name()
            })
    return result


# -*- coding: utf-8 -*-
"""
Submit.
"""
import json

import pymel.core as pm
from ciocore import CONFIG
import urlparse


def about_conductor():
    version = pm.moduleInfo(version=True, moduleName="conductor")
    definition = pm.moduleInfo(definition=True, moduleName="conductor")
    path = pm.moduleInfo(path=True, moduleName="conductor")

    result = """
Conductor for Maya

        Module version: {}
   Module install path: {}
Module definition file: {}

LICENSE INFORMATION

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Copyright © 2020, Conductor Technologies.
    """.format(
        version, path, definition
    )

    return result


def layout_form(form, text, main_layout, *buttons):
    form.attachForm(text, "left", 2)
    form.attachForm(text, "right", 2)
    form.attachForm(text, "top", 2)
    form.attachNone(text, "bottom")

    form.attachForm(main_layout, "left", 2)
    form.attachForm(main_layout, "right", 2)
    form.attachControl(main_layout, "top", 2, text)
    form.attachControl(main_layout, "bottom", 2, buttons[0])

    form.attachForm(buttons[0], "left", 2)
    form.attachNone(buttons[0], "top")
    form.attachForm(buttons[0], "bottom", 2)

    if len(buttons) == 1:
        form.attachForm(buttons[0], "right", 2)
    else:  # 2
        form.attachPosition(buttons[0], "right", 2, 50)

        form.attachPosition(buttons[1], "left", 2, 50)
        form.attachForm(buttons[1], "right", 2)
        form.attachNone(buttons[1], "top")
        form.attachForm(buttons[1], "bottom", 2)


def show_as_json(data, **kw):
    title = kw.get("title", "Json Window")
    indent = kw.get("indent", 2)
    sort_keys = kw.get("sort_keys", True)
    result_json = json.dumps(data, indent=indent, sort_keys=sort_keys)
    pm.window(width=600, height=800, title=title)
    pm.frameLayout(cll=False, lv=False)
    pm.scrollField(text=result_json, editable=False, wordWrap=False)
    pm.showWindow()

def _dismiss():
    pm.layoutDialog(dismiss="abort")

def _okay():
    pm.layoutDialog(dismiss="okay")

 

def submission_responses_layout(responses):
    form = pm.setParent(q=True)
    pm.formLayout(form, edit=True, width=300)
    text = pm.text(label="Links to {}".format(CONFIG["auth_url"]))
    b1 = pm.button(label="Close", command=pm.Callback(_okay))
    scroll = pm.scrollLayout(bv=True)

    pm.setParent("..")
    layout_form(form, text, scroll, b1)
    pm.setParent(scroll)
    pm.columnLayout(adjustableColumn=True, columnAttach=("both", 5), rowSpacing=10)
    for success_uri in [
        response["response"]["uri"].replace("jobs", "job")
        for response in responses
        if response.get("code") <= 201
    ]:
        job_url = urlparse.urljoin(CONFIG["auth_url"], success_uri)
        label = '<a href="{}"><font  color=#ec6a17 size=4>{}</font></a>'.format(
            job_url, job_url
        )
        pm.text(hl=True, label=label)

    failed_submissions = [r for r in responses if r.get("code") > 201]
    num_failed = len(failed_submissions)
    if num_failed:
        pm.separator(style="single")
        pm.text(hl=True, label="{:d} failed submissions".format(num_failed))
        pm.separator(style="single")
        for failed_submission in failed_submissions:
            text = failed_submission["response"]
            pm.text(label=text, align="left", wordWrap=False)

    pm.setParent(form)


def about_layout():
    form = pm.setParent(q=True)
    pm.formLayout(form, edit=True, width=300)
    heading = pm.text(label="Release Info and License")
    b1 = pm.button(label="Close", command=pm.Callback(_dismiss))
    frame = pm.frameLayout(
        bv=True, lv=False, cll=False, cl=False, width=700, height=500
    )
    pm.setParent("..")
    layout_form(form, heading, frame, b1)
    pm.setParent(frame)

    pm.scrollField(editable=False, wordWrap=True, text=about_conductor())

    pm.setParent(form)


def show_submission_responses(responses):
    return pm.layoutDialog(
        ui=pm.Callback(submission_responses_layout, responses), title="Render Response"
    )


def show_about():
    return pm.layoutDialog(ui=pm.Callback(about_layout), title="About Conductor-Maya")
